/*
  infoServ.cpp

  Creates a basic HTTP server for record list requests and other data
  (files, env vars, etc).

  /help? URL provides brief explanation of client HTTP interface.

  see doc/infoServ.html for more complete documentation

  2015-12-14  J. Priller  fixed off-by-1 bug in infoServ_JSONEscape w/control chars,
                          (wrote one byte past end of string)
  2015-06-03  J. Priller  fixed mem leak
                          (failure to call dbFinishEntry in
                          infoServ_pvs_requestHandler::doRecordJSONOutput)
  2015-04-09  J. Priller  ifdef for R3.14.15 macCreateHandle inputs
  2015-04-03  J. Priller  HTTP header added to built-in URL handler replies
  2015-03-16  J. Priller  added documentation
  2015-03-01  J. Priller  fix \u00xx JSON escaping
  2014-11-27  J. Priller  original version
*/

#define OUR_VERSION_MAJOR       1
#define OUR_VERSION_MINOR       1
#define OUR_VERSION_REVISION    0
#define OUR_VERSION_PATCHLEVEL  3

#define _STR_HELPER(x) #x
#define _STR(x) _STR_HELPER(x)
#define OUR_VERSION_STRING "infoServ " \
                           _STR(OUR_VERSION_MAJOR)    "." \
                           _STR(OUR_VERSION_MINOR)    "." \
                           _STR(OUR_VERSION_REVISION) "." \
                           _STR(OUR_VERSION_PATCHLEVEL)

#include "infoServ.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <osiSock.h>
#include <epicsThread.h>
#include <epicsMutex.h>
#include <dbStaticLib.h>
#include <dbDefs.h>
#include <dbAccessDefs.h>
#include <dbCommon.h>
#include <registryFunction.h>
#include <epicsExport.h>
#include <iocsh.h>
#include <macLib.h>
#include <epicsVersion.h>
#include <initHooks.h>

#ifdef vxWorks
  #include <wtx.h>
#endif

#ifdef _WIN32
  #define PATH_CHAR '\\'
#else
  #define PATH_CHAR '/'
#endif

#define EPICS_VERSION_LONG ( EPICS_VERSION*10000 + EPICS_REVISION*100 + EPICS_MODIFICATION )
#if (EPICS_VERSION_LONG >= 31408)
  #include <epicsStdioRedirect.h>
  #include <epicsExit.h>
  #define HAS_AT_EXIT
#endif

#ifndef epicsError
  #define epicsError -1
#endif
#ifndef epicsOk
  #define epicsOk 0
#endif

#ifndef DBRN_FLAGS_HASALIAS
  #define dbIsAlias(n) (false)
#endif

#define SERVER_THREAD_NAME  "infoServ_server"
#define CLIENT_THREAD_NAME  "infoServ_client"

#define INTERNAL_BUFFER_SIZE  1024

typedef struct {
  ELLNODE node;
  infoServ_HTTPClient *client;
} ELL_clientNode;

typedef struct {
  ELLNODE node;
  infoServ_requestHandler *handler;
} ELL_urlHandlerNode;

//----------------------------------------------------------------------
// internal class definitions
//----------------------------------------------------------------------
class infoServ_pvs_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_pvs_requestHandler();
    virtual ~infoServ_pvs_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
    void doRecordJSONOutput(SOCKET sock, infoServ_recordData *record, ELLLIST *fieldNames);
};

//----------------------------------------------------------------------
class infoServ_env_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_env_requestHandler();
    virtual ~infoServ_env_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
};

//----------------------------------------------------------------------
class infoServ_file_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_file_requestHandler();
    virtual ~infoServ_file_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
  protected:
    char *helpRows;
};

//----------------------------------------------------------------------
class infoServ_fileexists_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_fileexists_requestHandler();
    virtual ~infoServ_fileexists_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
  protected:
    char *helpRows;
};

//----------------------------------------------------------------------
class infoServ_filepaths_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_filepaths_requestHandler();
    virtual ~infoServ_filepaths_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
};

//----------------------------------------------------------------------
class infoServ_startup_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_startup_requestHandler();
    virtual ~infoServ_startup_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
};

//----------------------------------------------------------------------
class infoServ_help_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_help_requestHandler();
    virtual ~infoServ_help_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
};

//----------------------------------------------------------------------
class infoServ_info_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_info_requestHandler();
    virtual ~infoServ_info_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *infoString();
    virtual const char *htmlHelpTableRows();
  protected:
    char buildDateTime[32];
};

//----------------------------------------------------------------------
class infoServ_busy_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_busy_requestHandler();
    virtual ~infoServ_busy_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
};

//----------------------------------------------------------------------
class infoServ_root_requestHandler : public infoServ_requestHandler
{
  public:
    infoServ_root_requestHandler();
    virtual ~infoServ_root_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
};

//----------------------------------------------------------------------

// internal global variables
static infoServ_HTTPServer *infoServ_server = NULL;
static initHookState highestInitHookState = initHookAtBeginning;

// config params settable via infoServSetParam iocsh function
static int infoServ_debugLevel        = 0;
static int infoServ_port              = DEFAULT_INFOSERV_PORT;
static int infoServ_maxClients        = 10;
static int infoServ_selectTimeoutSec  = 0;          // 1/10 sec
static int infoServ_selectTimeoutUSec = 100000;     //   "   "
static int infoServ_clientLingerSec   = 5;
static int infoServ_clientTimeoutSec  = 10;
static int infoServ_threadRecycleSec  = 30;

static void serverPortChange(void);

typedef struct {
  const char *paramName;
  int *valuePointer;
  void (*changeFunc)(void);
} setParamItem;

static setParamItem setParamTable[] = {
  // 12345678901234567890
  { "DebugLevel",        &infoServ_debugLevel,        NULL             },
  { "Port",              &infoServ_port,              serverPortChange },
  { "MaxClients",        &infoServ_maxClients,        NULL             },
  { "SelectTimeoutSec",  &infoServ_selectTimeoutSec,  NULL             },
  { "SelectTimeoutUSec", &infoServ_selectTimeoutUSec, NULL             },
  { "ClientLingerSec",   &infoServ_clientLingerSec,   NULL             },
  { "ClientTimeoutSec",  &infoServ_clientTimeoutSec,  NULL             },
  { "ThreadRecycleSec",  &infoServ_threadRecycleSec,  NULL             },
  { NULL,                NULL,                        NULL             }
};

// list of printable chars needing json escaping: double-quote, slash
// non-printable characters are detected otherwise
static const char *infoServ_JSONEscape_chars = "\x22\x5c";

#ifndef HEXCHAR
#define HEXCHAR(n) (((n) <= 9) ? ((n) + '0') : ((n) + 'A' - 10))
#endif

//----------------------------------------------------------------------
// internal routines
//----------------------------------------------------------------------
#define TNOW_MAX 32
static char *tnow(char *storage)
{
  epicsTimeStamp now;
  epicsTimeGetCurrent(&now);
  epicsTimeToStrftime(storage,TNOW_MAX-1,"%Y/%m/%d %H:%M:%S",&now);
  return(storage);
}
//----------------------------------------------------------------------
bool checkAlloc(void *ptr, const char *file, int line)
{
  char tt[TNOW_MAX];

  if (ptr) return(true);
  printf("%s  infoServ: memory allocation failed! file %s line %d\n",tnow(tt),file,line);
  return(false);
}
#define MEM_ALLOC_CHECK(p) checkAlloc(p,__FILE__,__LINE__)
//----------------------------------------------------------------------
static unsigned char unhex(char ch)
{
  if ((ch >= '0') && (ch <= '9')) return( ch - '0' );
  if ((ch >= 'A') && (ch <= 'F')) return( ch - 'A' + 10 );
  if ((ch >= 'a') && (ch <= 'f')) return( ch - 'a' + 10 );
  return(0);
}
//----------------------------------------------------------------------
#ifdef HAS_AT_EXIT
static void infoServAtExit(void *arg)
{
  if (infoServ_server) {
    delete infoServ_server;
    infoServ_server = NULL;
  }
}
#endif
//----------------------------------------------------------------------
static void infoServStartHookFunc(initHookState state)
{
  if (state > highestInitHookState) highestInitHookState = state;
}
//----------------------------------------------------------------------
static void serverPortChange(void)
{
  // called by infoServSetParam(Port,x)
  // if server has started, send it new port number so it can restart listen
  if (infoServ_server) infoServ_server->changePort(infoServ_port);
}
//----------------------------------------------------------------------
static bool requireServerObject(void)
{
  if (! infoServ_server) {
    infoServ_server = new infoServ_HTTPServer(infoServ_port);
    MEM_ALLOC_CHECK(infoServ_server);
  }
  return( (infoServ_server) ? true : false );
}
//----------------------------------------------------------------------
static bool startsWith(const char *str, const char *match, char **tail)
{
  int kk = strlen(match);
  if (strncmp(str,match,kk) == 0) {
    (*tail) = (char *) str + kk;
    return(true);
  }
  (*tail) = NULL;
  return(false);
}
//----------------------------------------------------------------------
static void showBuffer(void *ptr, int size)
{
  int si, ei, ii, hi, ci;
  char *blankline =
  (char *)"                       -                                          ";
  char line[80];

  unsigned char *buff = (unsigned char *) ptr;
  printf("%d byte(s):\n",size);
  si = 0;
  do {
    strcpy(line,blankline);
    ei = si + 15;
    if (ei >= size) ei = size - 1;
    for (ii = si, hi = 0, ci = 50; ii <= ei; ii++, hi+=3, ci++) {
      unsigned char ch = buff[ii];
      line[hi]   = HEXCHAR(ch / 16);
      line[hi+1] = HEXCHAR(ch % 16);
      line[ci]   = ((ch < ' ') || (ch > '~')) ? '.' : ch;
    }
    printf("\t%s\n",line);
    si = ei + 1;
  } while (si < size);
}
//----------------------------------------------------------------------
static char *dupString(const char *str, int extra = 0)
{
  char *p;
  if (str) {
    p = new char[ strlen(str) + 1 + extra ];
    if (! MEM_ALLOC_CHECK(p)) return(NULL);
    strcpy(p,str);
  } else {
    // passed a NULL, create a null terminated string
    p = new char[ 1 + extra ];
    if (! MEM_ALLOC_CHECK(p)) return(NULL);
    p[0] = 0;
  }
  return(p);
}
//----------------------------------------------------------------------
static ELL_textNode *createTextNode(const char *text)
{
  ELL_textNode *textNode = new ELL_textNode;
  if (! MEM_ALLOC_CHECK(textNode)) return(NULL);

  memset(textNode,0,sizeof(ELL_textNode));
  textNode->text = dupString(text);
  if (! textNode->text) {
    delete textNode;
    return(NULL);
  }

  return(textNode);
}
//----------------------------------------------------------------------
static bool addToTextList(ELLLIST *L, const char *text)
{
  if (! L) return(false);
  ELL_textNode *textNode = createTextNode(text);
  if (! textNode) return(false);
  ellAdd(L,&textNode->node);
  return(true);
}
//----------------------------------------------------------------------
static char *allocFindAndReplace(const char *origcontent, const char *findtxt, const char *repltxt)
{
  // count number of replacements needed
  int numrep = 0;
  char *pc = (char *) origcontent;
  while (pc) {
    pc = strstr(pc,findtxt);
    if (pc) {
      numrep++;
      pc += strlen(findtxt);
    }
  }

  // allocate space for new text
  // will be a little large because we don't consider repllen - findlen
  // (if we did, what it repl was shorter than find?)
  int findlen = strlen(findtxt);
  int repllen = strlen(repltxt);
  int numc = strlen(origcontent) + (repllen * numrep) + 1;
  char *content = new char[numc];
  if (! MEM_ALLOC_CHECK(content)) return(NULL);

  // do replacements on content
  strcpy(content,origcontent);
  do {
    char *pc = strstr(content,findtxt);
    if (! pc) break;
    memmove(pc+repllen,pc+findlen,strlen(pc+findlen)+1);
    memcpy(pc,repltxt,repllen);
  } while (true);

  return(content);
}
//----------------------------------------------------------------------
static char *fetchEnvValue(const char *name, char *value, int maxlen)
{
  // create macro evaluation handle
  MAC_HANDLE *hndl;
#if (EPICS_VERSION_LONG >= 31415)
  static const char *pairs[] = { (char *)"", (char *)"environ", NULL, NULL };  // borrowed from macEnv.c
#else
  static char *pairs[] = { (char *)"", (char *)"environ", NULL, NULL };  // borrowed from macEnv.c
#endif
  long stat = macCreateHandle(&hndl,pairs);
  if (stat != epicsOk) return(NULL);

  // suppress warnings
  macSuppressWarning(hndl,1);

  // evaluate
  char tmp[INTERNAL_BUFFER_SIZE];
  strcpy(tmp,"$(");
  strcat(tmp,name);
  strcat(tmp,")");
  int numc = macExpandString(hndl,tmp,value,maxlen);

  // delete handle
  macDeleteHandle(hndl);

  return( (numc >= 0) ? value : NULL );
}
//----------------------------------------------------------------------
// routines visible to other modules
//----------------------------------------------------------------------
infoServ_HTTPServer *infoServ_getServer(void)
{
  return(infoServ_server);
}
//----------------------------------------------------------------------
void infoServ_parseCommaSeparated(const char *str, ELLLIST *L, const char *endChars)
{
  if (! L) return;
  char *tcopy = dupString(str);
  int si = 0;
  int ei = 0;
  while (true) {
    if (tcopy[ei] == 0) break;
    if (strchr(endChars,tcopy[ei])) break;
    if ((tcopy[ei] < ' ') || (tcopy[ei] > '~')) break;
    if (tcopy[ei] == ',')  {
      tcopy[ei] = 0;
      if (! addToTextList(L,tcopy+si)) return;
      si = ei + 1;
      ei = si;
    } else {
      ei++;
    }
  }
  if (ei != si) {
    tcopy[ei] = 0;
    if (! addToTextList(L,tcopy+si)) return;
  }
  delete [] tcopy;
}
//----------------------------------------------------------------------
void infoServ_copyTextList(ELLLIST *targ, ELLLIST *src)
{
  if ((! targ) || (! src)) return;
  ELL_textNode *textNode = (ELL_textNode *) ellFirst(src);
  while (textNode) {
    if (textNode->text) addToTextList(targ,textNode->text);
    textNode = (ELL_textNode *) ellNext(&textNode->node);
  }
}
//----------------------------------------------------------------------
void infoServ_destroyTextList(ELLLIST *L)
{
  if (! L) return;
  ELL_textNode *textNode;
  while ((textNode = (ELL_textNode *) ellFirst(L))) {
    if (textNode->text) delete [] textNode->text;
    ellDelete(L,&textNode->node);
    delete textNode;
  }
}
//----------------------------------------------------------------------
void infoServ_buildRecordList(ELLLIST *L)
{
  DBENTRY dbentry;
  DBENTRY *pdbentry = &dbentry;

  /*
    There's no way to simply get all the aliases that refer to a record, so we
    have to make one pass creating a list of all records and then do another
    pass going through each alias telling records what aliases point to them.
  */

  // pass 1: fill in list with records
  dbInitEntry(pdbbase,pdbentry);
  int stat = dbFirstRecordType(pdbentry);
  while (! stat) {
    stat = dbFirstRecord(pdbentry);
    while (stat == 0) {
      // create item for this record
      ELL_recordDataNode *recordNode = new ELL_recordDataNode;
      if (MEM_ALLOC_CHECK(recordNode)) {
        recordNode->record = new infoServ_recordData(pdbentry);
        if (! MEM_ALLOC_CHECK(recordNode->record)) {
          delete recordNode;
        } else {
          ellAdd(L,&recordNode->node);
        }
      }
      // next record
      stat = dbNextRecord(pdbentry);
    }
    stat = dbNextRecordType(pdbentry);
  }
  dbFinishEntry(pdbentry);

  // pass 2: fill in aliases-of for entries
  ELL_recordDataNode *recordNode = (ELL_recordDataNode *) ellFirst(L);
  while (recordNode) {
    if ((recordNode->record) && (recordNode->record->isAlias())) {
      // locate item we're an alias to and inform it we're an alias of it
      ELL_recordDataNode *huntNode = (ELL_recordDataNode *) ellFirst(L);
      while (huntNode) {
        if ((huntNode->record)
         && (strcmp(recordNode->record->aliasTo(),huntNode->record->recordName()) == 0)) {
          huntNode->record->addAliasOf(recordNode->record->recordName());
          break;
        }
        huntNode = (ELL_recordDataNode *) ellNext(&huntNode->node);
      }
    }
    recordNode = (ELL_recordDataNode *) ellNext(&recordNode->node);
  }
}
//----------------------------------------------------------------------
void infoServ_destroyRecordList(ELLLIST *L)
{
  ELL_recordDataNode *recordNode;
  while ((recordNode = (ELL_recordDataNode *) ellFirst(L))) {
    if (recordNode->record) delete recordNode->record;
    ellDelete(L,&recordNode->node);
    delete recordNode;
  }
}
//----------------------------------------------------------------------
bool infoServ_sortList(ELLLIST *L, ELL_COMPARE_FUNC func)
{
  if ((! L) || (! func)) return(false);

  // get node count
  int currCount = ellCount(L);
  if (currCount <= 1) return(true);

  // allocate space for list's node pointers
  ELLNODE **pPtrList = new ELLNODE* [currCount];
  if (! MEM_ALLOC_CHECK(pPtrList)) return(false);

  // copy list's pointers into allocated array
  ELLNODE *pNode = ellFirst(L);
  int idx = 0;
  while (pNode) {
    pPtrList[idx] = pNode;
    idx++;
    pNode = ellNext(pNode);
  }

  // do the sort
  int gap = currCount / 2;
  do {
    for (int ii = 0; ii < (currCount - gap); ii++) {
      int X = ii;
      do {
        if (func(pPtrList[X],pPtrList[X+gap]) <= 0) break;
        ELLNODE *tmp = pPtrList[X];
        pPtrList[X] = pPtrList[X+gap];
        pPtrList[X+gap] = tmp;
        X -= gap;
      } while (X >= 0);
    }
    gap = gap / 2;
  } while (gap != 0);

  // rebuild list from newly sorted pointer array
  ellInit(L);
  for (int ii = 0; ii < currCount; ii++) {
    ellAdd(L,pPtrList[ii]);
  }

  // clean up and we're done
  delete [] pPtrList;
  return(true);
}
//----------------------------------------------------------------------
int infoServ_sortCompareTextNodes(ELLNODE *n1, ELLNODE *n2)
{
  ELL_textNode *t1 = (ELL_textNode *) n1;
  ELL_textNode *t2 = (ELL_textNode *) n2;
  return( strcmp(t1->text,t2->text) );
}
//----------------------------------------------------------------------
int infoServ_sortCompareRecordNodeNames(ELLNODE *n1, ELLNODE *n2)
{
  ELL_recordDataNode *r1 = (ELL_recordDataNode *) n1;
  ELL_recordDataNode *r2 = (ELL_recordDataNode *) n2;
  if ((! r1) || (! r2) || (! r1->record) || (! r2->record)) return(0);
  return( strcmp(r1->record->recordName(),r2->record->recordName()) );
}
//----------------------------------------------------------------------
bool infoServ_registerUrlHandler(infoServ_requestHandler *handler)
{
  if (! requireServerObject()) return(false);
  return( infoServ_server->registerUrlHandler(handler) );
}
//----------------------------------------------------------------------
bool infoServ_unregisterUrlHandler(infoServ_requestHandler *handler)
{
  if (! requireServerObject()) return(false);
  return( infoServ_server->unregisterUrlHandler(handler) );
}
//----------------------------------------------------------------------
void infoServ_htmlUnescape(char *buff)
// un-converts HTML escape sequences ('%20' -> ' ', etc)
{
  int ii, jj, N, numc;

  numc = strlen(buff);
  ii = 0;
  jj = 0;
  while (ii < numc) {
    if (buff[ii] != '%') {
      buff[jj] = buff[ii];
      ii++;
      jj++;
    } else {
      N = unhex(buff[ii+1]);
      N <<= 4;
      N |= unhex(buff[ii+2]);
      buff[jj] = N;
      ii += 3;
      jj++;
    }
  }
  buff[jj] = 0;
}
//----------------------------------------------------------------------
void infoServ_socketSendString(SOCKET sock, const char *str)
{
  send(sock,str,strlen(str),0);
}
//----------------------------------------------------------------------
void infoServ_sendHTTPText(SOCKET sock, const char *statcode, const char *contents)
{
  char tmp[INTERNAL_BUFFER_SIZE];

  // send the HTTP message header
  infoServ_socketSendString(sock,"HTTP/1.1 ");
  infoServ_socketSendString(sock,statcode);
  infoServ_socketSendString(sock,"\r\n");
  infoServ_socketSendString(sock,"Content-Encoding: text\r\n");
  sprintf(tmp,"Content-Length: %d\r\n",strlen(contents));
  infoServ_socketSendString(sock,tmp);
  infoServ_socketSendString(sock,"Content-Type: text/html; charset=iso-8859-1\r\n");
  infoServ_socketSendString(sock,"Connection: close\r\n");
  infoServ_socketSendString(sock,"\r\n");

  // send contents
  infoServ_socketSendString(sock,contents);
}
//----------------------------------------------------------------------
void infoServ_send404Error(SOCKET sock, const char *badurl)
{
  static const char *err404content =  \
     "<!DOCTYPE html>\r\n"                                                          \
     "<html>\r\n"                                                                   \
     "<head><title>404 Not Found</title></head>\r\n"                                \
     "<body>\r\n"                                                                   \
     "<h1>404 Not Found</h1>\r\n"                                                   \
     "The requested URL !BADURL! could not be processed by this server\r\n"         \
     "<br>\r\n"                                                                     \
     "<br>\r\n"                                                                     \
     "<hr>\r\n"                                                                     \
     "<address>" OUR_VERSION_STRING "</address>\r\n"                                \
     "</body>\r\n"                                                                  \
     "</html>";

  // generate content
  char *content = allocFindAndReplace(err404content,"!BADURL!",badurl);
  if (! content) return;

  // send it and we're done
  infoServ_sendHTTPText(sock,"404 Not Found",content);
  delete [] content;
}
//----------------------------------------------------------------------
void infoServ_send503Error(SOCKET sock)
{
  static const char *busycontent =                              \
     "<!DOCTYPE html>\r\n"                                      \
     "<html>\r\n"                                               \
     "<head><title>503 Service Unavailable</title></head>\r\n"  \
     "<body>\r\n"                                               \
     "<h1>503 Service Unavailable</h1>\r\n"                     \
     "No free connections available!\r\n"                       \
     "<br>\r\n"                                                 \
     "<br>\r\n"                                                 \
     "<hr>\r\n"                                                 \
    "<address>" OUR_VERSION_STRING "</address>\r\n"             \
     "</body>\r\n"                                              \
     "</html>";

  infoServ_sendHTTPText(sock,"503 Service Unavailable",busycontent);
}

//----------------------------------------------------------------------
bool infoServ_needJSONEscape(const char *buff)
{
  for (int ii = 0; buff[ii] != 0; ii++) {
    if (strchr(infoServ_JSONEscape_chars,buff[ii]))
      return(true);
    else if ((buff[ii] < ' ') || (buff[ii] > '~'))
      return(true);
  }
  return(false);
}
//----------------------------------------------------------------------
int infoServ_JSONEscapeSize(const char *buff)
// calculates what size string will expand to if infoServ_JSONEscape() called
{
  int newsize = strlen(buff);
  for (int ii = 0; buff[ii] != 0; ii++) {
    if (strchr(infoServ_JSONEscape_chars,buff[ii]))
      newsize++;  // needs a backslash escape char
    else if ((buff[ii] < ' ') || (buff[ii] > '~'))
      newsize += 5;  // needs to be \u00NN
  }
  return(newsize);
}
//----------------------------------------------------------------------
void infoServ_JSONEscape(char *buff)
// performs escape-sequencing for a single-quoted json text string
// we escape printable chars in 'infoServ_JSONEscape_chars'
// and all non-printable chars
{
  int ii;

  ii = 0;
  while (buff[ii] != 0) {
    if (strchr(infoServ_JSONEscape_chars,buff[ii])) {
      memmove(buff+ii+1,buff+ii,strlen(buff+ii)+1);
      buff[ii] = '\\';
      ii += 2;
    } else if ((buff[ii] < ' ') || (buff[ii] > '~')) {
      int ch = buff[ii];
      if (ch < 0) ch += 0x0100;
      memmove(buff+ii+6,buff+ii+1,strlen(buff+ii+1)+1);
      buff[ii]   = '\\';
      buff[ii+1] = 'u';
      buff[ii+2] = '0';
      buff[ii+3] = '0';
      buff[ii+4] = HEXCHAR(ch / 16);
      buff[ii+5] = HEXCHAR(ch % 16);
      ii += 6;
    } else {
      ii++;
    }
  }
}
//----------------------------------------------------------------------
void infoServ_socketSendJSONText(SOCKET sock, const char *text)
{
  if (! infoServ_needJSONEscape(text)) {
    infoServ_socketSendString(sock,"\"");
    infoServ_socketSendString(sock,text);
    infoServ_socketSendString(sock,"\"");
  } else {
    char *pc = new char[ infoServ_JSONEscapeSize(text) + 1 ];
    if (! MEM_ALLOC_CHECK(pc)) return;
    strcpy(pc,text);
    infoServ_JSONEscape(pc);
    infoServ_socketSendString(sock,"\"");
    infoServ_socketSendString(sock,pc);
    infoServ_socketSendString(sock,"\"");
    delete [] pc;
  }
}
//----------------------------------------------------------------------
void infoServ_socketSendJSONFieldName(SOCKET sock, const char *fieldname)
{
  infoServ_socketSendJSONText(sock,fieldname);
  infoServ_socketSendString(sock,": ");
}
//----------------------------------------------------------------------
void infoServ_socketSendJSONPairText(SOCKET sock, const char *fieldname, const char *fieldvalue)
{
  // send field name
  infoServ_socketSendJSONFieldName(sock,fieldname);

  // send field value
  if (! fieldvalue) {
    // no field value, use null
    infoServ_socketSendString(sock,"null");
  } else {
    infoServ_socketSendJSONText(sock,fieldvalue);
  }
}
//----------------------------------------------------------------------
void infoServ_socketSendJSONPairLong(SOCKET sock, const char *fieldname, long fieldvalue)
{
  // send field name
  infoServ_socketSendJSONFieldName(sock,fieldname);

  // send field value
  char tmp[64];
  sprintf(tmp,"%ld",fieldvalue);
  infoServ_socketSendString(sock,tmp);
}
//----------------------------------------------------------------------
void infoServ_socketSendJSONPairBool(SOCKET sock, const char *fieldname, bool fieldvalue)
{
  // send field name
  infoServ_socketSendJSONFieldName(sock,fieldname);

  // send field value
  if (fieldvalue)
    infoServ_socketSendString(sock,"true");
  else
    infoServ_socketSendString(sock,"false");
}
//----------------------------------------------------------------------
void infoServ_socketSendJSONPairDouble(SOCKET sock, const char *fieldname, const char *format, double fieldvalue)
{
  // send field name
  infoServ_socketSendJSONFieldName(sock,fieldname);

  // send field value
  char tmp[64];
  sprintf(tmp,format,fieldvalue);
  infoServ_socketSendString(sock,tmp);
}
//----------------------------------------------------------------------
const char *infoServ_getStartupFile(char *buffer, int maxlen)
{
#ifdef vxWorks
  if (strlen(sysBootParams.startupScript) < maxlen)
    strcpy(buffer,sysBootParams.startupScript);
  else {
    strncpy(buffer,sysBootParams.startupScript,maxlen-1);
    buffer[maxlen-1] = 0;
  }
  return(buffer);
#else
  // this will work for R3.14.12.2
  // also for NSCL's R3.14.9 IOCs where main iocsh app sets it
  return( fetchEnvValue("IOCSH_STARTUP_SCRIPT",buffer,maxlen) );
#endif
}
//----------------------------------------------------------------------
// iocsh callable functions
//----------------------------------------------------------------------
void infoServStart(void)
{
  // make sure server object exists
  requireServerObject();
  if (! infoServ_server) {
    printf("infoServ: no server object, init must have failed\n");
    return;
  }

  // start the server if it isn't already
  if (! infoServ_server->started()) infoServ_server->start();
}
//----------------------------------------------------------------------
void infoServStop(void)
{
  if (infoServ_server) infoServ_server->stop();
}
//----------------------------------------------------------------------
int infoServSetParam(const char *param, int value)
{
  if ((param == NULL) || (param[0] == 0)) {
    // no argument, show current values
    printf("parameter name        value\n");
    printf("--------------------  -----\n");
    for (int ii = 0; setParamTable[ii].paramName != NULL; ii++) {
      printf("%-20s  %d\n",setParamTable[ii].paramName,(*setParamTable[ii].valuePointer));
    }
  } else {
    bool found = false;
    for (int ii = 0; setParamTable[ii].paramName != NULL; ii++) {
      if (strcmp(setParamTable[ii].paramName,param) == 0) {
        found = true;
        int oldVal = (*setParamTable[ii].valuePointer);
        (*setParamTable[ii].valuePointer) = value;
        if (setParamTable[ii].changeFunc) (*setParamTable[ii].changeFunc)();
        printf("infoServ set %s to %d (was %d)\n",param,value,oldVal);
        break;
      }
    }
    if (! found) printf("infoServ error: no such settable parameter '%s'\n",param);
  }
  return(epicsOk);
}
//----------------------------------------------------------------------
void infoServAllowFilePath(const char *newpath)
{
  requireServerObject();

  if (! infoServ_server) {
    printf("infoServ: no server object, init must have failed\n");
    return;
  }

  if ((newpath) && (strlen(newpath) > 0))
    infoServ_server->addAllowedFilePath(newpath);

  // always show allowed file paths list
  infoServ_server->showAllowedFilePaths();
}
//----------------------------------------------------------------------
void infoServReport(int level)
{
  if (! infoServ_server) {
    printf("infoServ doesn't appear to be started\n");
    return;
  }
  infoServ_server->report(level);
}
//----------------------------------------------------------------------
// CLASSES
//----------------------------------------------------------------------
// class infoServ_recordData
//----------------------------------------------------------------------
infoServ_recordData::infoServ_recordData(DBENTRY *pdbentry)
{
  long stat;

  name     = NULL;
  rtyp     = NULL;
  alias_to = NULL;
  ellInit(&aliases_of);

  // get name and type
  name = dupString( dbGetRecordName(pdbentry) );
  rtyp = dupString( dbGetRecordTypeName(pdbentry) );

  // get alias target if we're an alias
  if (dbIsAlias(pdbentry)) {
    stat = dbFindField(pdbentry,"NAME");
    if (stat == 0) alias_to = dupString( dbGetString(pdbentry) );
  }
}
//----------------------------------------------------------------------
infoServ_recordData::~infoServ_recordData(void)
{
  if (name)     { delete [] name;     name     = NULL; }
  if (rtyp)     { delete [] rtyp;     rtyp     = NULL; }
  if (alias_to) { delete [] alias_to; alias_to = NULL; }
  infoServ_destroyTextList(&aliases_of);
}
//----------------------------------------------------------------------
const char *infoServ_recordData::recordName(void)
{
  return(name);
}
//----------------------------------------------------------------------
const char *infoServ_recordData::recordType(void)
{
  return(rtyp);
}
//----------------------------------------------------------------------
const char *infoServ_recordData::aliasTo(void)
{
  return(alias_to);
}
//----------------------------------------------------------------------
bool infoServ_recordData::isAlias(void)
{
  return( (alias_to) ? true : false );
}
//----------------------------------------------------------------------
ELLLIST *infoServ_recordData::aliasesOf(void)
{
  return(&aliases_of);
}
//----------------------------------------------------------------------
void infoServ_recordData::addAliasOf(const char *other)
{
  addToTextList(&aliases_of,other);
}
//----------------------------------------------------------------------
// class infoServ_requestHandler
//----------------------------------------------------------------------
infoServ_requestHandler::infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_requestHandler::~infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
void infoServ_requestHandler::serverStarted()
{
  // nothing to do here
}
//----------------------------------------------------------------------
void infoServ_requestHandler::serverStopped()
{
  // nothing to do here
}
//----------------------------------------------------------------------
void infoServ_requestHandler::report(int level)
{
  // nothing to do here
}
//----------------------------------------------------------------------
const char *infoServ_requestHandler::infoString()
{
  return(NULL);
}
//----------------------------------------------------------------------
const char *infoServ_requestHandler::htmlHelpTableRows()
{
  return(NULL);
}
//----------------------------------------------------------------------
void infoServ_requestHandler::sendTextHeader(SOCKET sock, const char *extraHeaderLines)
{
  infoServ_socketSendString(sock,"HTTP/1.1 200 OK\r\n");
  infoServ_socketSendString(sock,"Content-Encoding: text\r\n");
  infoServ_socketSendString(sock,"Content-Type: text/plain; charset=iso-8859-1\r\n");
  if (extraHeaderLines) infoServ_socketSendString(sock,extraHeaderLines);
  infoServ_socketSendString(sock,"Connection: close\r\n");
  infoServ_socketSendString(sock,"\r\n");
}
//----------------------------------------------------------------------
void infoServ_requestHandler::sendJSONHeader(SOCKET sock, const char *extraHeaderLines)
{
  infoServ_socketSendString(sock,"HTTP/1.1 200 OK\r\n");
  infoServ_socketSendString(sock,"Content-Encoding: text\r\n");
  infoServ_socketSendString(sock,"Content-Type: application/json; charset=iso-8859-1\r\n");
  if (extraHeaderLines) infoServ_socketSendString(sock,extraHeaderLines);
  infoServ_socketSendString(sock,"Connection: close\r\n");
  infoServ_socketSendString(sock,"\r\n");
}
//----------------------------------------------------------------------
// class infoServ_pvs_requestHandler
//----------------------------------------------------------------------
infoServ_pvs_requestHandler::infoServ_pvs_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_pvs_requestHandler::~infoServ_pvs_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_pvs_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/pvs?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_pvs_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  ELLLIST entries;
  ELLLIST fields;

  // parse request url tail (fields=x,y,etc)
  ellInit(&fields);
  if ((strncmp(urlText+5,"fields=",7) == 0)) {  // skip /pvs?
    infoServ_parseCommaSeparated(urlText+5+7,&fields);
  }

  // create list with record names, types and aliases
  ellInit(&entries);
  infoServ_buildRecordList(&entries);

  // sort the list by record name
  infoServ_sortList(&entries,infoServ_sortCompareRecordNodeNames);

  // output HTTP header
  sendJSONHeader(sock);

  // output contents header
  infoServ_socketSendString(sock,"{ \"pvs\": {\r\n");

  // output results
  ELL_recordDataNode *recordNode = (ELL_recordDataNode *) ellFirst(&entries);
  while (recordNode) {
    if (recordNode->record) doRecordJSONOutput(sock, recordNode->record, &fields);
    recordNode = (ELL_recordDataNode *) ellNext(&recordNode->node);
    if (recordNode) infoServ_socketSendString(sock,",");
    infoServ_socketSendString(sock,"\r\n");
  }

  // output footer
  infoServ_socketSendString(sock,"} }\r\n");

  // dispose of allocated memory
  infoServ_destroyTextList(&fields);
  infoServ_destroyRecordList(&entries);

  // done
  return(true);
}
//----------------------------------------------------------------------
void infoServ_pvs_requestHandler::doRecordJSONOutput(SOCKET sock, infoServ_recordData *record, ELLLIST *fieldNames)
{
  long stat;
  DBENTRY dbentry;
  DBENTRY *pdbentry = &dbentry;

  // header w/name & type
  infoServ_socketSendJSONText(sock,record->recordName());
  infoServ_socketSendString(sock,": { ");
  infoServ_socketSendJSONPairText(sock,"type",record->recordType());

  // get record DB entry we'll need for gathering other data
  dbInitEntry(pdbbase,pdbentry);
  stat = dbFindRecord(pdbentry,record->recordName());
  if (stat != epicsOk) {
    // this should never fail
    // but if it somehow does we need to send terminator for name/type info above
    infoServ_socketSendString(sock," }");
    return;
  }

  // if we're an alias to something, include that
  if (record->aliasTo()) {
    infoServ_socketSendString(sock,", ");
    infoServ_socketSendJSONPairText(sock,"is_alias_to",record->aliasTo());
  }

  // fields requested, if any
  if ((fieldNames) && (ellCount(fieldNames) > 0)) {
    infoServ_socketSendString(sock,", \"fields\": { ");
    ELL_textNode *pText = (ELL_textNode *) ellFirst(fieldNames);
    while (pText) {
      stat = dbFindField(pdbentry,pText->text);
      if (stat == 0)
        infoServ_socketSendJSONPairText(sock,pText->text,dbGetString(pdbentry));
      else
        infoServ_socketSendJSONPairText(sock,pText->text,NULL);
      pText = (ELL_textNode *) ellNext(&pText->node);
      if (pText) infoServ_socketSendString(sock,", ");
    }
    infoServ_socketSendString(sock," }");
  }

  // aliases
  if (ellCount(record->aliasesOf()) > 0) {
    infoServ_socketSendString(sock,", \"aliases\": [ ");
    ELL_textNode *pText = (ELL_textNode *) ellFirst(record->aliasesOf());
    while (pText) {
      infoServ_socketSendJSONText(sock,pText->text);
      pText = (ELL_textNode *) ellNext(&pText->node);
      if (pText) infoServ_socketSendString(sock,", ");
    }
    infoServ_socketSendString(sock," ]");
  }

  // send any info fields
  stat = dbFirstInfo(pdbentry);
  if (stat == 0) {
    infoServ_socketSendString(sock,", \"info\": {");
    while (stat == 0) {
      const char *iname = dbGetInfoName(pdbentry);
      const char *itext = dbGetInfoString(pdbentry);
      infoServ_socketSendJSONPairText(sock,iname,itext);
      stat = dbNextInfo(pdbentry);
      if (stat == 0) infoServ_socketSendString(sock,", ");
    }
    infoServ_socketSendString(sock," }");
  }

  // done with db entry
  dbFinishEntry(pdbentry);

  // footer
  infoServ_socketSendString(sock," }");
}
//----------------------------------------------------------------------
const char *infoServ_pvs_requestHandler::infoString()
{
  return("/pvs? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_pvs_requestHandler::htmlHelpTableRows()
{
  static const char *pvsHelp = \
    "<tr>\r\n"                                                                           \
    "  <td><b>/pvs?[fields=FIELD[,FIELD2...]]</b></td>\r\n"                              \
    "  <td>(JSON data) get list of records on IOC, and optionally field values</td>\r\n" \
    "  <td align=\"center\"><a href=\"/pvs?fields=ADEL,DESC\">test it!</a></td>\r\n"     \
    "</tr>\r\n";

  return(pvsHelp);
}
//----------------------------------------------------------------------
// class infoServ_env_requestHandler
//----------------------------------------------------------------------
infoServ_env_requestHandler::infoServ_env_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_env_requestHandler::~infoServ_env_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_env_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/env?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_env_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  ELLLIST envs;
  ellInit(&envs);
  infoServ_parseCommaSeparated(urlText+5,&envs);  // skip /env?

  // output HTTP header
  sendJSONHeader(sock);

  // output contentsheader
  infoServ_socketSendString(sock,"{ \"env\": {\r\n");

  // output env vars
  ELL_textNode *textNode = (ELL_textNode *) ellFirst(&envs);
  while (textNode) {
    char buff[INTERNAL_BUFFER_SIZE];
    infoServ_socketSendJSONPairText(sock,textNode->text,fetchEnvValue(textNode->text,buff,sizeof(buff)));
    textNode = (ELL_textNode *) ellNext(&textNode->node);
    if (textNode) infoServ_socketSendString(sock,",");
    infoServ_socketSendString(sock,"\r\n");
  }

  // output footer
  infoServ_socketSendString(sock,"} }\r\n");

  // clean up
  infoServ_destroyTextList(&envs);

  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_env_requestHandler::infoString()
{
  return("/env? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_env_requestHandler::htmlHelpTableRows()
{
  static const char *envHelp = \
    "<tr>\r\n"                                                                          \
    "  <td><b>/env?VARNAME[,VARNAME2...]</b></td>\r\n"                                  \
    "  <td>(JSON data) get values of IOC environment variables</td>\r\n"                \
    "  <td align=\"center\"><a href=\"/env?TOP,EPICS_BASE,ARCH\">test it!</a></td>\r\n" \
    "</tr>\r\n";

  return(envHelp);
}
//----------------------------------------------------------------------
// class infoServ_file_requestHandler
//----------------------------------------------------------------------
infoServ_file_requestHandler::infoServ_file_requestHandler()
  : infoServ_requestHandler()
{
  helpRows = NULL;  // generated when required
}
//----------------------------------------------------------------------
infoServ_file_requestHandler::~infoServ_file_requestHandler()
{
  if (helpRows) {
    delete [] helpRows;
    helpRows = NULL;
  }
}
//----------------------------------------------------------------------
bool infoServ_file_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/file?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_file_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  // need a file name specified
  const char *fileName = urlText + 6;
  if (strlen(fileName) == 0) return(false);

  // can't have .. in it
  if (strstr(fileName,"..") != NULL) return(false);

  // there must be a server object
  infoServ_HTTPServer *server = infoServ_getServer();
  if (! server) return(false);

  // must be an allowed file path
  if (! server->isAllowedFilePath(fileName)) return(false);

  // okay to read file if it exists (and we have read access to it obviously)
  FILE *fi = fopen(fileName,"r");
  if (fi == NULL) return(false);

  // allocate a buffer to read data into
  int buffchars = 10240;
  char *buff = new char[buffchars];
  if (! MEM_ALLOC_CHECK(buff)) {
    fclose(fi);
    return(false);
  }

  // output HTTP header
  sendTextHeader(sock);

  // send file data
  while (fgets(buff,buffchars-1,fi)) {
    infoServ_socketSendString(sock,buff);
  }
  fclose(fi);

  // we're done
  delete [] buff;
  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_file_requestHandler::infoString()
{
  return("/file? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_file_requestHandler::htmlHelpTableRows()
{
  static const char *fileHelp = \
    "<tr>\r\n"                                                                                         \
    "  <td><b>/file?FILENAME</b></td>\r\n"                                                             \
    "  <td>(TEXT) get contents of a given file<br>\r\n"                                                \
    "      absolute directory paths <b>ONLY</b><br>\r\n"                                               \
    "      <b>MUST</b> be subdir of $(TOP) or allowed by IOC call to infoServAllowFilePath()</td>\r\n" \
    "  <td align=\"center\"><a href=\"/file?!FILENAME!\">test it!</a></td>\r\n"                        \
    "</tr>\r\n";

  if (! helpRows) {
    char buff[INTERNAL_BUFFER_SIZE];
    const char *testFile = infoServ_getStartupFile(buff,sizeof(buff));
    if (! testFile) testFile = "/cannot/determine/startup/file";
    helpRows = allocFindAndReplace(fileHelp,"!FILENAME!",testFile);
    if (! helpRows) return(fileHelp);
  }

  return(helpRows);
}
//----------------------------------------------------------------------
// class infoServ_fileexists_requestHandler
//----------------------------------------------------------------------
infoServ_fileexists_requestHandler::infoServ_fileexists_requestHandler()
  : infoServ_requestHandler()
{
  helpRows = NULL;  // generated when required
}
//----------------------------------------------------------------------
infoServ_fileexists_requestHandler::~infoServ_fileexists_requestHandler()
{
  if (helpRows) {
    delete [] helpRows;
    helpRows = NULL;
  }
}
//----------------------------------------------------------------------
bool infoServ_fileexists_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/fileexists?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_fileexists_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  // need a file name specified
  const char *fileName = urlText + 12;  // get past /fileexists?
  if (strlen(fileName) == 0) return(false);

  // can't have .. in it
  if (strstr(fileName,"..") != NULL) return(false);

  // there must be a server object
  infoServ_HTTPServer *server = infoServ_getServer();
  if (! server) return(false);

  // must be an allowed file path
  if (! server->isAllowedFilePath(fileName)) return(false);

  // exists if we can open if for a read
  bool exists = false;
  FILE *fi = fopen(fileName,"r");
  if (fi != NULL) {
    fclose(fi);
    exists = true;
  }

  sendJSONHeader(sock);
  infoServ_socketSendString(sock,"{ \"fileexists\": {\r\n");
  infoServ_socketSendJSONPairBool(sock,fileName,exists);
  infoServ_socketSendString(sock," } }\r\n");

  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_fileexists_requestHandler::infoString()
{
  return("/fileexists? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_fileexists_requestHandler::htmlHelpTableRows()
{
  static const char *fileexistsHelp = \
    "<tr>\r\n"                                                                        \
    "  <td><b>/fileexists?FILENAME</b></td>\r\n"                                      \
    "  <td>(JSON) whether the given file exists, same limitations as /file?</td>\r\n" \
    "  <td align=\"center\"><a href=\"/fileexists?!FILENAME!\">test it!</a></td>\r\n" \
    "</tr>\r\n";

  if (! helpRows) {
    char buff[INTERNAL_BUFFER_SIZE];
    const char *testFile = infoServ_getStartupFile(buff,sizeof(buff));
    if (! testFile) testFile = "/cannot/determine/startup/file";
    helpRows = allocFindAndReplace(fileexistsHelp,"!FILENAME!",testFile);
    if (! helpRows) return(fileexistsHelp);
  }

  return(helpRows);
}
//----------------------------------------------------------------------
// class infoServ_filepaths_requestHandler
//----------------------------------------------------------------------
infoServ_filepaths_requestHandler::infoServ_filepaths_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_filepaths_requestHandler::~infoServ_filepaths_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_filepaths_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/filepaths?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_filepaths_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  // there must be a server object
  infoServ_HTTPServer *server = infoServ_getServer();
  if (! server) return(false);

  // get copy of server object's file paths
  ELLLIST allowedFilePathsList;
  ellInit(&allowedFilePathsList);
  server->getAllowedFilePaths(&allowedFilePathsList);

  sendJSONHeader(sock);
  infoServ_socketSendString(sock,"{ \"filepaths\": [\r\n");

  ELL_textNode *textNode = (ELL_textNode *) ellFirst(&allowedFilePathsList);
  while (textNode) {
    infoServ_socketSendJSONText(sock,textNode->text);
    textNode = (ELL_textNode *) ellNext(&textNode->node);
    if (textNode) infoServ_socketSendString(sock,",");
    infoServ_socketSendString(sock,"\r\n");
  }
  infoServ_socketSendString(sock,"] }\r\n");

  infoServ_destroyTextList(&allowedFilePathsList);

  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_filepaths_requestHandler::infoString()
{
  return("/filepaths? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_filepaths_requestHandler::htmlHelpTableRows()
{
  static const char *helpRows = \
    "<tr>\r\n"                                                                              \
    "  <td><b>/filepaths?</b></td>\r\n"                                                     \
    "  <td>(JSON data) get list of allowed file paths for /file? and /fileexists?</td>\r\n" \
    "  <td align=\"center\"><a href=\"/filepaths?\">test it!</a></td>\r\n"                  \
    "</tr>\r\n";

  return(helpRows);
}
//----------------------------------------------------------------------
// class infoServ_startup_requestHandler
//----------------------------------------------------------------------
infoServ_startup_requestHandler::infoServ_startup_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_startup_requestHandler::~infoServ_startup_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_startup_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/startup?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_startup_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  char buff[INTERNAL_BUFFER_SIZE];
  const char *startFile = infoServ_getStartupFile(buff,sizeof(buff));
  sendJSONHeader(sock);
  infoServ_socketSendString(sock,"{ \"startup\": {\r\n");
  infoServ_socketSendJSONPairText(sock,"file",startFile);
  infoServ_socketSendString(sock,",\r\n");
#ifdef vxWorks
  // see:
  // http://www.kumanov.com/docs/prog/VxWorks/vxworks/netguide/c-booting2.html
  infoServ_socketSendJSONPairText(sock,"boothost",sysBootParams.hostName);
#else
  infoServ_socketSendJSONPairText(sock,"boothost",NULL);
#endif
  infoServ_socketSendString(sock,"\r\n} }");

  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_startup_requestHandler::infoString()
{
  return("/startup? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_startup_requestHandler::htmlHelpTableRows()
{
  const char *helpRows = \
    "<tr>\r\n"                                                           \
    "  <td><b>/startup?</b></td>\r\n"                                    \
    "  <td>(JSON data) get IOC startup script info</td>\r\n"             \
    "  <td align=\"center\"><a href=\"/startup?\">test it!</a></td>\r\n" \
    "</tr>\r\n";

  return(helpRows);
}
//----------------------------------------------------------------------
// class infoServ_info_requestHandler
//----------------------------------------------------------------------
infoServ_info_requestHandler::infoServ_info_requestHandler()
  : infoServ_requestHandler()
{
  // convert DATE/TIME macros into a YYYY-mm-dd HH:MM:SS format
  const char *tod = __DATE__ " " __TIME__;
  const char *monthNames = "JanFebMarAprMayJunJulAugSepOctNovDec";
  char mtxt[4];
  memcpy(mtxt,tod,3);
  mtxt[3] = 0;
  int mnum = 0;
  const char *mptr = strstr(monthNames,mtxt);
  if (mptr) mnum = ((mptr - monthNames) / 3) + 1;

  strcpy(buildDateTime,"YYYY-mm-dd HH:MM:SS");
  memcpy(buildDateTime,tod+7,4);
  if (mnum > 0) {
    buildDateTime[5] = '0' + (mnum / 10);
    buildDateTime[6] = '0' + (mnum % 10);
  }
  buildDateTime[5] = '0' + (mnum / 10);
  buildDateTime[8] = (tod[4] == ' ') ? '0' : tod[4];
  buildDateTime[9] = tod[5];
  memcpy(buildDateTime+11,tod+12,8);
}
//----------------------------------------------------------------------
infoServ_info_requestHandler::~infoServ_info_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_info_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/info?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_info_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  // there must be a server object
  infoServ_HTTPServer *server = infoServ_getServer();
  if (! server) return(false);

  sendJSONHeader(sock);
  infoServ_socketSendString(sock,"{ \"info\": {\r\n");

  char sep[2];
  sep[0] = PATH_CHAR;
  sep[1] = 0;
  infoServ_socketSendString(sock,"  ");
  infoServ_socketSendJSONPairText(sock,"pathseparator",sep);
  infoServ_socketSendString(sock,",\r\n");

  infoServ_socketSendString(sock,"  ");
  infoServ_socketSendString(sock,"\"version\": {\r\n");
  infoServ_socketSendString(sock,"    ");
  infoServ_socketSendJSONPairLong(sock,"major",OUR_VERSION_MAJOR);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"minor",OUR_VERSION_MINOR);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"revision",OUR_VERSION_REVISION);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"patchlevel",OUR_VERSION_PATCHLEVEL);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairText(sock,"versionstring",OUR_VERSION_STRING);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairText(sock,"built",buildDateTime);
  infoServ_socketSendString(sock,"\r\n  },\r\n");

  infoServ_socketSendString(sock,"  ");
  infoServ_socketSendString(sock,"\"epicsversion\": {\r\n");
  infoServ_socketSendString(sock,"    ");
  infoServ_socketSendJSONPairLong(sock,"version",EPICS_VERSION);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"revision",EPICS_REVISION);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"modification",EPICS_MODIFICATION);
  infoServ_socketSendString(sock,",\r\n    ");
#ifdef EPICS_PATCH_LEVEL
  infoServ_socketSendJSONPairLong(sock,"patchlevel",EPICS_PATCH_LEVEL);
  infoServ_socketSendString(sock,",\r\n    ");
#else
  #ifdef EPICS_UPDATE_LEVEL
  infoServ_socketSendJSONPairLong(sock,"updatelevel",EPICS_UPDATE_LEVEL);
  infoServ_socketSendString(sock,",\r\n    ");
  #endif
#endif
  infoServ_socketSendJSONPairText(sock,"versionstring",EPICS_VERSION_STRING);
  infoServ_socketSendString(sock,"\r\n  },\r\n");

  ELLLIST infoStrings;
  ellInit(&infoStrings);
  server->getHandlerInfoStrings(&infoStrings);
  infoServ_socketSendString(sock,"  ");
  infoServ_socketSendString(sock,"\"urlhandlers\": [\r\n");
  ELL_textNode *textNode = (ELL_textNode *) ellFirst(&infoStrings);
  while (textNode) {
    if (textNode->text) {
      infoServ_socketSendString(sock,"    ");
      infoServ_socketSendJSONText(sock,textNode->text);
    }
    textNode = (ELL_textNode *) ellNext(&textNode->node);
    if (textNode) infoServ_socketSendString(sock,",");
    infoServ_socketSendString(sock,"\r\n");
  }
  infoServ_socketSendString(sock,"  ],\r\n");
  infoServ_destroyTextList(&infoStrings);

  infoServ_serverStats stats;
  server->getServerStats(&stats);
  infoServ_socketSendString(sock,"  ");
  infoServ_socketSendString(sock,"\"serverstats\": {\r\n");
  infoServ_socketSendString(sock,"    ");
  infoServ_socketSendJSONPairLong(sock,"upseconds",stats.upSeconds);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"currentclientcount",stats.currentClientCount);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"highestclientcount",stats.highestClientCount);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"createdclientcount",stats.createdClientCount);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"deletedclientcount",stats.deletedClientCount);
  infoServ_socketSendString(sock,",\r\n    ");
  infoServ_socketSendJSONPairLong(sock,"recycledclientcount",stats.recycledClientCount);
  infoServ_socketSendString(sock,"\r\n  },\r\n");

  infoServ_socketSendString(sock,"  ");
  infoServ_socketSendString(sock,"\"params\": {\r\n");
  for (int ii = 0; setParamTable[ii].paramName != NULL; ii++) {
    infoServ_socketSendString(sock,"    ");
    infoServ_socketSendJSONPairLong(sock,setParamTable[ii].paramName,(*setParamTable[ii].valuePointer));
    bool last = (setParamTable[ii+1].paramName == NULL);
    if (! last) infoServ_socketSendString(sock,",");
    infoServ_socketSendString(sock,"\r\n");
  }
  infoServ_socketSendString(sock,"  }\r\n");

  infoServ_socketSendString(sock,"} }");

  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_info_requestHandler::infoString()
{
  return("/info? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_info_requestHandler::htmlHelpTableRows()
{
  static const char *helpRows = \
    "<tr>\r\n"                                                        \
    "  <td><b>/info?</b></td>\r\n"                                    \
    "  <td>(JSON data) server information</td>\r\n"                   \
    "  <td align=\"center\"><a href=\"/info?\">test it!</a></td>\r\n" \
    "</tr>\r\n";

  return(helpRows);
}
//----------------------------------------------------------------------
// class infoServ_help_requestHandler
//----------------------------------------------------------------------
infoServ_help_requestHandler::infoServ_help_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_help_requestHandler::~infoServ_help_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_help_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/help?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_help_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  static const char *helpheader = \
     "<!DOCTYPE html>\r\n"                                 \
     "<html>\r\n"                                          \
     "<head><title>infoServ IOC server</title></head>\r\n" \
     "<body>\r\n"                                          \
     "<h1>infoServ IOC information server</h1>\r\n"        \
     "<h3>URLs supported:</h3>\r\n"                        \
     "<table cellpadding='4' border='1'>\r\n";

  static const char *helpfooter = \
     "</table>\r\n"                                  \
     "<br>\r\n"                                      \
     "<hr>\r\n"                                      \
     "<address>" OUR_VERSION_STRING "</address>\r\n" \
     "</body>\r\n"                                   \
     "</html>";

  // there must be a server object
  infoServ_HTTPServer *server = infoServ_getServer();
  if (! server) return(false);

  // get help rows from server, which knows all the registered URL handlers
  // and will query them for the table contents they want displayed
  ELLLIST handlerTableRows;
  ellInit(&handlerTableRows);
  server->getHandlerHelpTableRows(&handlerTableRows);

  // calculate size of buffer we need for complete content
  int bufferSize = strlen(helpheader);
  ELL_textNode *textNode = (ELL_textNode *) ellFirst(&handlerTableRows);
  while (textNode) {
    if (textNode->text) bufferSize += strlen(textNode->text);
    textNode = (ELL_textNode *) ellNext(&textNode->node);
  }
  bufferSize += strlen(helpfooter) + 1;

  // generate content, send it, delete it
  char *content = new char [bufferSize];
  if (MEM_ALLOC_CHECK(content)) {
    strcpy(content,helpheader);
    ELL_textNode *textNode = (ELL_textNode *) ellFirst(&handlerTableRows);
    while (textNode) {
      if (textNode->text) strcat(content,textNode->text);
      textNode = (ELL_textNode *) ellNext(&textNode->node);
    }
    strcat(content,helpfooter);
    infoServ_sendHTTPText(sock,"200 OK",content);
    delete [] content;
  }

  infoServ_destroyTextList(&handlerTableRows);

  return(true);
}
//----------------------------------------------------------------------
const char *infoServ_help_requestHandler::infoString()
{
  return("/help? (built-in)");
}
//----------------------------------------------------------------------
const char *infoServ_help_requestHandler::htmlHelpTableRows()
{
  const char *helpRows = \
    "<tr>\r\n"                                       \
    "  <td><b>/help?</b></td>\r\n"                   \
    "  <td>this page</td>\r\n"                       \
    "  <td align=\"center\">you're here...</td>\r\n" \
    "</tr>\r\n";

  return(helpRows);
}
//----------------------------------------------------------------------
// class infoServ_busy_requestHandler
//----------------------------------------------------------------------
infoServ_busy_requestHandler::infoServ_busy_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_busy_requestHandler::~infoServ_busy_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_busy_requestHandler::canHandleUrl(const char *urlText)
{
  char *pc;
  return( startsWith(urlText,"/busy?",&pc) );
}
//----------------------------------------------------------------------
bool infoServ_busy_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  infoServ_send503Error(sock);
  return(true);
}
//----------------------------------------------------------------------
// class infoServ_root_requestHandler
//----------------------------------------------------------------------
infoServ_root_requestHandler::infoServ_root_requestHandler()
  : infoServ_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
infoServ_root_requestHandler::~infoServ_root_requestHandler()
{
  // nothing to do here
}
//----------------------------------------------------------------------
bool infoServ_root_requestHandler::canHandleUrl(const char *urlText)
{
  return( (strcmp(urlText,"/") == 0) ? true : false );
}
//----------------------------------------------------------------------
bool infoServ_root_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  static const char *rootcontent =                                     \
     "<!DOCTYPE html>\r\n"                                             \
     "<html>\r\n"                                                      \
     "<head><title>infoServ IOC information server</title></head>\r\n" \
     "<body>\r\n"                                                      \
     "<h1>infoServ IOC information server</h1>\r\n"                    \
     "URL help is available at <a href=\"/help?\">/help?</a>\r\n"      \
     "<br>\r\n"                                                        \
     "<br>\r\n"                                                        \
     "<hr>\r\n"                                                        \
     "<address>" OUR_VERSION_STRING "</address>\r\n"                   \
     "</body>\r\n"                                                     \
     "</html>";
  infoServ_sendHTTPText(sock,"200 OK",rootcontent);
  return(true);
}
//----------------------------------------------------------------------
// class infoServ_HTTPServer
//----------------------------------------------------------------------
infoServ_HTTPServer::infoServ_HTTPServer(short port)
{
  pThread          = NULL;
  isStarted        = false;
  stopThread       = false;
  needCreateSocket = true;
  ourPort          = port;
  ourSock          = INVALID_SOCKET;  // socket created in run()
  listsMutexId     = epicsMutexCreate();
  startTime        = epicsTime::getCurrent();
  memset(&serverStats,0,sizeof(serverStats));
  ellInit(&clientList);
  ellInit(&allowedFilePathsList);
  ellInit(&urlHandlerList);

  // $(TOP) is always an allowed file path if it's defined
  char buff[INTERNAL_BUFFER_SIZE];
  const char *top = fetchEnvValue("TOP",buff,sizeof(buff));
  if (top) {
    addAllowedFilePath(top);
#ifdef vxWorks
    // for VxWorks, boothost:/$TOP is allowed is boothost not already start of $TOP
    if (strlen(sysBootParams.hostName) > 0) {
      char *pc;
      if (! startsWith(top,sysBootParams.hostName,&pc)) {
        char *boottop = new char [ strlen(sysBootParams.hostName) + 1 + strlen(top) + 1 ];
        if (MEM_ALLOC_CHECK(boottop)) {
          strcpy(boottop,sysBootParams.hostName);
          strcat(boottop,":");
          strcat(boottop,top);
          addAllowedFilePath(boottop);
          delete [] boottop;
        }
      }
    }
#endif
  }

  // register default URL handler objects
  registerUrlHandler( new infoServ_help_requestHandler() );
  registerUrlHandler( new infoServ_info_requestHandler() );
  registerUrlHandler( new infoServ_pvs_requestHandler() );
  registerUrlHandler( new infoServ_env_requestHandler() );
  registerUrlHandler( new infoServ_file_requestHandler() );
  registerUrlHandler( new infoServ_fileexists_requestHandler() );
  registerUrlHandler( new infoServ_filepaths_requestHandler() );
  registerUrlHandler( new infoServ_startup_requestHandler() );
  registerUrlHandler( new infoServ_busy_requestHandler() );
  registerUrlHandler( new infoServ_root_requestHandler() );
}
//----------------------------------------------------------------------
infoServ_HTTPServer::~infoServ_HTTPServer()
{
  // tell server to stop
  stop();

  // destroy socket
  if (ourSock != INVALID_SOCKET) {
    epicsSocketDestroy(ourSock);
    ourSock = INVALID_SOCKET;
  }

  // remove any remaining clients
  ELL_clientNode *clientNode;
  while ((clientNode = (ELL_clientNode *) ellFirst(&clientList))) {
    if (clientNode->client) {
      delete clientNode->client;
      serverStats.deletedClientCount++;
    }
    ellDelete(&clientList,&clientNode->node);
    delete clientNode;
  }

  // delete allowed file paths
  infoServ_destroyTextList(&allowedFilePathsList);

  // delete url handler list
  // we also delete any remaining handler objects
  ELL_urlHandlerNode *handlerNode;
  while ((handlerNode = (ELL_urlHandlerNode *) ellFirst(&urlHandlerList))) {
    if (handlerNode->handler) {
      delete handlerNode->handler;
    }
    ellDelete(&urlHandlerList,&handlerNode->node);
    delete handlerNode;
  }

  // delete mutex
  epicsMutexDestroy(listsMutexId);
  listsMutexId = 0;
}
//----------------------------------------------------------------------
bool infoServ_HTTPServer::started()
{
  return(isStarted);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::start()
{
  if (! isStarted) {
    if (! pThread) {
      // create thread object
      pThread = new epicsThread(*this,
                                SERVER_THREAD_NAME,
                                epicsThreadGetStackSize(epicsThreadStackBig),
                                epicsThreadPriorityLow);
      if (! MEM_ALLOC_CHECK(pThread)) return;
    }
    stopThread = false;
    pThread->start();
    isStarted = true;
  }
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::stop()
{
  char tt[TNOW_MAX];

  // stop and delete thread, if we have one
  if (pThread) {
    if (isStarted) {
      printf("%s infoServ: stopping server thread...\n",tnow(tt));
      stopThread = true;
      if (! pThread->exitWait(5.0)) {
        printf("%s infoServ: timed out waiting for server thread to quit\n",tnow(tt));
      }
    }
    delete pThread;
    pThread = NULL;
  }

  // wait for clients to quit
  int waited = 0;
  while (cleanup() > 0) {
    if (waited > 10) {
      printf("%s infoServ: timed out waiting for clients to quit\n",tnow(tt));
      break;
    }
    epicsThreadSleep(1.0);
    ++waited;
  }

  isStarted = false;
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::run()
{
  fd_set fds;
  struct timeval timeout;
  struct sockaddr_in sin;
  osiSocklen_t sin_len;
  char desc[256];
  char tt[TNOW_MAX];
  ELL_urlHandlerNode *handlerNode;
  long lastUpSec;
  int select_stat;
  int fdset_stat;

  // inform url handler objects we've started
  epicsMutexLock(listsMutexId);
  handlerNode = (ELL_urlHandlerNode *) ellFirst(&urlHandlerList);
  while (handlerNode) {
    if (handlerNode->handler) handlerNode->handler->serverStarted();
    handlerNode = (ELL_urlHandlerNode *) ellNext(&handlerNode->node);
  }
  epicsMutexUnlock(listsMutexId);

  stopThread = false;
  lastUpSec = -1;
  while (! stopThread) {
    // update up-seconds
    serverStats.upSeconds = (long) (epicsTime::getCurrent() - startTime);
    if (serverStats.upSeconds != lastUpSec) {
      lastUpSec = serverStats.upSeconds;
      if (infoServ_debugLevel >= 10) {
        printf("%s infoServ: up seconds: %ld\n",tnow(tt),serverStats.upSeconds);
      }
    }

    // if IOC startup not yet advanced enough, do nothing
    // no sense accepting clients we can't yet give any record info to...
    if (highestInitHookState < initHookAtEnd) {
      if (infoServ_debugLevel >= 20) {
        printf("%s infoServ: highestInitHookState: %d\n",tnow(tt),highestInitHookState);
      }
      epicsThreadSleep(1.0);
      continue;
    }

    // clean up any completed/expired clients
    cleanup();

    // (re)create our listening socket if we need to
    if (needCreateSocket) {
      if (infoServ_debugLevel >= 2) {
        printf("%s infoServ: needCreateSocket\n",tnow(tt));
      }
      if (ourSock != INVALID_SOCKET) {
        epicsSocketDestroy(ourSock);
        ourSock = INVALID_SOCKET;
      }
      struct sockaddr_in local;
      ourSock = epicsSocketCreate(AF_INET, SOCK_STREAM, 0);
      local.sin_family = AF_INET;
      local.sin_port = htons(ourPort);
#ifdef WIN32
      local.sin_addr.S_un.S_addr = htonl(INADDR_ANY);
#else
      local.sin_addr.s_addr = htonl(INADDR_ANY);
#endif
#ifndef vxWorks
      // don't need REUSE in VxWorks, whole OS restarts on IOC reboot
      int val = 1;
      setsockopt(ourSock, SOL_SOCKET, SO_REUSEADDR, (const char *)&val, sizeof(val));
#endif
      if (bind(ourSock, (struct sockaddr *)&local, sizeof local) != 0) {
        // bind failed, delete socket and we'll try again in a few seconds
        printf("%s infoServ: bind port %d FAILED\n",tnow(tt),ourPort);
        epicsSocketDestroy(ourSock);
        ourSock = INVALID_SOCKET;
        epicsThreadSleep(5.0);
        continue;
      }
      listen(ourSock,5);
      needCreateSocket = false;
      if (infoServ_debugLevel > 0)
        printf("%s infoServ: infoServ_HTTPServer listening on port %d\n",tnow(tt),ourPort);
    }

    // check for new connections
    FD_ZERO(&fds);
    FD_SET(ourSock, &fds);
    timeout.tv_sec  = infoServ_selectTimeoutSec;
    timeout.tv_usec = infoServ_selectTimeoutUSec;

    select_stat = select(ourSock+1, &fds, 0, 0, &timeout);
    if (select_stat == 1) {
      fdset_stat = FD_ISSET(ourSock, &fds);
      if (fdset_stat) {
        sin_len = sizeof(sin);
        SOCKET clientsock = epicsSocketAccept(ourSock, (struct sockaddr *) &sin, &sin_len);
        if (clientsock != INVALID_SOCKET) {
          // set client socket linger
          struct linger linger;
          linger.l_onoff = 1;
          linger.l_linger = infoServ_clientLingerSec;
          setsockopt(clientsock, SOL_SOCKET, SO_LINGER, (char *)&linger, sizeof(linger));

          // get description text for socket (ip:port)
          ipAddrToDottedIP((const sockaddr_in*)&sin,desc,sizeof(desc));

          // try to assign client/socket to a client thread (created or recycled)
          bool ok = assignClientThread(clientsock,desc);
          if (ok) {
            if (infoServ_debugLevel > 0)
              printf("%s infoServ: infoServ_HTTPServer %s client connected\n",tnow(tt),desc);
          } else {
            // no client threads available, send 503 code
            infoServ_send503Error(clientsock);
            shutdown(clientsock,SHUT_RDWR);
            epicsSocketDestroy(clientsock);
            printf("%s infoServ ERROR: infoServ_HTTPServer reached %d concurrent clients\n",
                   tnow(tt),infoServ_maxClients);
          }

        } // socket okay
      } else {
        if (infoServ_debugLevel >= 30) {
          printf("%s infoServ: FD_ISSET returned %d\n",tnow(tt),fdset_stat);
        }
      }
    } else {
      if (infoServ_debugLevel >= 30) {
        printf("%s infoServ: select() returned %d\n",tnow(tt),select_stat);
      }
    }

  } // while go

  // delete our socket
  if (ourSock != INVALID_SOCKET) {
    epicsSocketDestroy(ourSock);
    ourSock = INVALID_SOCKET;
  }
  needCreateSocket = true;

  // inform url handler objects we've stopped
  epicsMutexLock(listsMutexId);
  handlerNode = (ELL_urlHandlerNode *) ellFirst(&urlHandlerList);
  while (handlerNode) {
    if (handlerNode->handler) handlerNode->handler->serverStopped();
    handlerNode = (ELL_urlHandlerNode *) ellNext(&handlerNode->node);
  }
  epicsMutexUnlock(listsMutexId);

  if (infoServ_debugLevel > 0) printf("%s infoServ: infoServ_HTTPServer exiting\n",tnow(tt));
}
//----------------------------------------------------------------------
bool infoServ_HTTPServer::assignClientThread(SOCKET clientsock, const char *desc)
{
  bool retstat = false;

  epicsMutexLock(listsMutexId);

  // first see if we have a client/thread we can recycle
  bool foundRecycleClient = false;
  ELL_clientNode *clientNode = (ELL_clientNode *) ellFirst(&clientList);
  while (clientNode) {
    if ((clientNode->client) && (clientNode->client->canRecycle())) {
      foundRecycleClient = clientNode->client->recycle(clientsock,desc);
      if (foundRecycleClient) {
        retstat = true;
        serverStats.recycledClientCount++;
        break;
      }
    }
    clientNode = (ELL_clientNode *) ellNext(&clientNode->node);
  }

  if (! foundRecycleClient) {
    // couldn't find a client/thread to recycle, need to make a new one (if there's room)
    if (serverStats.currentClientCount < infoServ_maxClients) {
      // client can be added, create one and start its thread
      ELL_clientNode *clientNode = new ELL_clientNode;
      if (! MEM_ALLOC_CHECK(clientNode))
        retstat = false;
      else {
        clientNode->client = new infoServ_HTTPClient(this, clientsock, desc);
        if (! MEM_ALLOC_CHECK(clientNode->client)) {
          delete clientNode;
          retstat = false;
        } else {
          serverStats.createdClientCount++;
          ellAdd(&clientList,&clientNode->node);
          clientNode->client->start();
          retstat = true;
        }
      }
    } else {
      // no more clients allowed, send 503 code
      retstat = false;
    }
  }

  // update client stats
  serverStats.currentClientCount = ellCount(&clientList);
  if (serverStats.currentClientCount > serverStats.highestClientCount)
    serverStats.highestClientCount = serverStats.currentClientCount;

  epicsMutexUnlock(listsMutexId);

  return(retstat);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::handleUrl(SOCKET sock, const char *url)
{
  //
  // locate a registered URL handler that claims it can handle this URL
  //
  // we roll through registered handlers backwards so that user-registered
  // handlers can override our built-in ones
  //
  epicsMutexLock(listsMutexId);
  infoServ_requestHandler *handlerFound = NULL;
  ELL_urlHandlerNode *handlerNode = (ELL_urlHandlerNode *) ellLast(&urlHandlerList);
  while (handlerNode) {
    if ((handlerNode->handler) && (handlerNode->handler->canHandleUrl(url))) {
      handlerFound = handlerNode->handler;
      break;
    }
    handlerNode = (ELL_urlHandlerNode *) ellPrevious(&handlerNode->node);
  }
  epicsMutexUnlock(listsMutexId);

  // if no handler found, send 404 error
  if (! handlerFound) {
    infoServ_send404Error(sock,url);
    return;
  }

  // let handler try to handle it
  // if that fails send a 404 error
  bool handled = handlerFound->handleUrl(sock,url);
  if (! handled) infoServ_send404Error(sock,url);
}
//----------------------------------------------------------------------
bool infoServ_HTTPServer::registerUrlHandler(infoServ_requestHandler *handler)
{
  epicsMutexLock(listsMutexId);

  ELL_urlHandlerNode *handlerNode = new ELL_urlHandlerNode;
  if (MEM_ALLOC_CHECK(handlerNode)) {
    handlerNode->handler = handler;
    ellAdd(&urlHandlerList,&handlerNode->node);
  }

  epicsMutexUnlock(listsMutexId);
  return(true);
}
//----------------------------------------------------------------------
bool infoServ_HTTPServer::unregisterUrlHandler(infoServ_requestHandler *handler)
{
  epicsMutexLock(listsMutexId);

  ELL_urlHandlerNode *handlerNode = (ELL_urlHandlerNode *) ellFirst(&urlHandlerList);
  while (handlerNode) {
    if ((handlerNode->handler) && (handlerNode->handler == handler))
      break;
    else
      handlerNode = (ELL_urlHandlerNode *) ellNext(&handlerNode->node);
  }

  bool found = (handlerNode) ? true : false;
  if (handlerNode) {
    /*
      We found a match, remove item from the url handler list and delete the
      linked-list item it was in

      We *DO NOT* delete the infoServ_requestHandler object, if somebody is
      calling us to unregister then it's up to them to delete the object
    */
    ellDelete(&urlHandlerList,&handlerNode->node);
    delete handlerNode;
  }

  epicsMutexUnlock(listsMutexId);

  return(found);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::report(int level)
{
  epicsMutexLock(listsMutexId);

  printf("\n%s\n",OUR_VERSION_STRING);
  printf("server state : %s\n",(isStarted) ? "started" : "stopped");
  if (isStarted) {
    printf("up seconds   : %ld\n",serverStats.upSeconds);
    printf("client stats:\n");
    printf("  current client count  : %ld\n",serverStats.currentClientCount);
    printf("  highest client count  : %ld\n",serverStats.highestClientCount);
    printf("  created client count  : %ld\n",serverStats.createdClientCount);
    printf("  recycled client count : %ld\n",serverStats.recycledClientCount);
    printf("  deleted client count  : %ld\n",serverStats.deletedClientCount);
  }

  ELL_clientNode *clientNode = (ELL_clientNode *) ellFirst(&clientList);
  while (clientNode) {
    if (clientNode->client) clientNode->client->report(level);
    clientNode = (ELL_clientNode *) ellNext(&clientNode->node);
  }

  ELL_urlHandlerNode *handlerNode = (ELL_urlHandlerNode *) ellFirst(&urlHandlerList);
  while (handlerNode) {
    if (handlerNode->handler) handlerNode->handler->report(level);
    handlerNode = (ELL_urlHandlerNode *) ellNext(&handlerNode->node);
  }

  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::getServerStats(infoServ_serverStats *stats)
{
  epicsMutexLock(listsMutexId);
  memcpy(stats,&serverStats,sizeof(infoServ_serverStats));
  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::changePort(short newport)
{
  if (newport == ourPort) {
    char tt[TNOW_MAX];
    printf("%s infoServ: server port is already %d\n",tnow(tt),newport);
  } else {
    // set new port
    ourPort = newport;
    // flag for run() thread to re-create socket
    needCreateSocket = true;
  }
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::addAllowedFilePath(const char *newpath)
{
  // skip if we already have it
  if (isAllowedFilePath(newpath)) return;

  epicsMutexLock(listsMutexId);

  // get a copy we're sure is terminated by path char (/ or \)
  char *tmp = dupString(newpath,1);
  if (tmp) {
    int kk = strlen(tmp);
    if (tmp[kk-1] != PATH_CHAR) {
      tmp[kk]   = PATH_CHAR;
      tmp[kk+1] = 0;
    }

    // create a new text node and add to list
    ELL_textNode *textNode = createTextNode(tmp);
    if (textNode) ellAdd(&allowedFilePathsList,&textNode->node);

    // delete temporary storage
    delete [] tmp;
  }

  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::showAllowedFilePaths(void)
{
  epicsMutexLock(listsMutexId);

  printf("infoServ allowed file paths:\n");
  ELL_textNode *textNode = (ELL_textNode *) ellFirst(&allowedFilePathsList);
  while (textNode) {
    if (textNode->text) printf("  '%s'\n",textNode->text);
    textNode = (ELL_textNode *) ellNext(&textNode->node);
  }

  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
bool infoServ_HTTPServer::isAllowedFilePath(const char *filename)
{
  epicsMutexLock(listsMutexId);

  bool found = false;
  ELL_textNode *textNode = (ELL_textNode *) ellFirst(&allowedFilePathsList);
  while (textNode) {
    if (textNode->text) {
      if (strncmp(filename,textNode->text,strlen(textNode->text)) == 0) {
        found = true;
        break;
      }
    }
    textNode = (ELL_textNode *) ellNext(&textNode->node);
  }

  epicsMutexUnlock(listsMutexId);

  return(found);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::getAllowedFilePaths(ELLLIST *L)
{
  epicsMutexLock(listsMutexId);
  infoServ_copyTextList(L,&allowedFilePathsList);
  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::getHandlerHelpTableRows(ELLLIST *L)
{
  epicsMutexLock(listsMutexId);

  ELL_urlHandlerNode *handlerNode = (ELL_urlHandlerNode *) ellFirst(&urlHandlerList);
  while (handlerNode) {
    if (handlerNode->handler) {
      const char *pc = handlerNode->handler->htmlHelpTableRows();
      if (pc) addToTextList(L,pc);
    }
    handlerNode = (ELL_urlHandlerNode *) ellNext(&handlerNode->node);
  }

  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPServer::getHandlerInfoStrings(ELLLIST *L)
{
  epicsMutexLock(listsMutexId);

  // go through the order backwards
  // as that's how url-handling works
  // so that user-installed handlers can override built-in ones
  ELL_urlHandlerNode *handlerNode = (ELL_urlHandlerNode *) ellLast(&urlHandlerList);
  while (handlerNode) {
    if (handlerNode->handler) {
      const char *pc = handlerNode->handler->infoString();
      if (pc) addToTextList(L,pc);
    }
    handlerNode = (ELL_urlHandlerNode *) ellPrevious(&handlerNode->node);
  }

  epicsMutexUnlock(listsMutexId);
}
//----------------------------------------------------------------------
int infoServ_HTTPServer::cleanup()
{
  int live_clients = 0;

  epicsMutexLock(listsMutexId);

  ELL_clientNode *clientNode = (ELL_clientNode *) ellFirst(&clientList);
  while (clientNode) {
    bool remove = false;
    if (! clientNode->client) {
      // null pointer erroneously stored somehow, just remove
      remove = true;
    } else if (clientNode->client->canCleanup()) {
      remove = true;
    } else {
      live_clients++;
    }
    ELL_clientNode *nextClient = (ELL_clientNode *) ellNext(&clientNode->node);
    if (remove) {
      if (clientNode->client) delete clientNode->client;
      ellDelete(&clientList,&clientNode->node);
      delete clientNode;
      serverStats.deletedClientCount++;
    }
    clientNode = nextClient;
  }

  serverStats.currentClientCount = live_clients;

  epicsMutexUnlock(listsMutexId);

  return(live_clients);
}
//----------------------------------------------------------------------
// class infoServ_HTTPClient
//----------------------------------------------------------------------
infoServ_HTTPClient::infoServ_HTTPClient(infoServ_HTTPServer *server, SOCKET socket, const char *desc)
   : thread(*this,
            CLIENT_THREAD_NAME,
            epicsThreadGetStackSize(epicsThreadStackBig),
            epicsThreadPriorityLow)
{
  ourServer = server;
  ourSock   = INVALID_SOCKET;
  ourDesc   = NULL;
  mutexId   = epicsMutexCreate();
  changeState(stateInit);
  initRequest(socket,desc);
}
//----------------------------------------------------------------------
infoServ_HTTPClient::~infoServ_HTTPClient()
{
  char tt[TNOW_MAX];

  if (getState() != stateThreadExited) {
    const char *pc = ourDesc;
    if (! pc) pc = "[unknown]";
    printf("%s infoServ: client %s forced to shut down\n",tnow(tt),pc);
  }

  if (ourSock != INVALID_SOCKET) {
    epicsSocketDestroy(ourSock);
    ourSock = INVALID_SOCKET;
  }

  if (ourDesc) {
    delete [] ourDesc;
    ourDesc = NULL;
  }

  // delete mutex
  epicsMutexDestroy(mutexId);
  mutexId = 0;

  if (infoServ_debugLevel > 0)
    printf("%s infoServ: infoServ_HTTPClient thread deleted\n",tnow(tt));
}
//----------------------------------------------------------------------
void infoServ_HTTPClient::report(int level)
{
  epicsMutexLock(mutexId);
  const char *desc = (ourDesc) ? ourDesc : "[unknown]";
  printf("client thread\n");
  printf("  last request  : %s\n",lastRequest);
  printf("  received from : %s\n",desc);
  printf("  current state : %s\n",getStateText());
  epicsMutexUnlock(mutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPClient::initRequest(SOCKET socket, const char *desc)
// called by constructor and recycle()
{
  epicsMutexLock(mutexId);

  if (ourSock != INVALID_SOCKET) epicsSocketDestroy(ourSock);
  ourSock = socket;

  if (ourDesc != NULL) delete [] ourDesc;
  ourDesc = dupString(desc);

  requestStartTime = epicsTime::getCurrent();
  input[0]         = 0;
  lastRequest[0]   = 0;

  changeState(stateDoingRequest);

  epicsMutexUnlock(mutexId);
}
//----------------------------------------------------------------------
void infoServ_HTTPClient::changeState(clientState newState)
{
  epicsMutexLock(mutexId);
  ourState = newState;
  epicsMutexUnlock(mutexId);
}
//----------------------------------------------------------------------
infoServ_HTTPClient::clientState infoServ_HTTPClient::getState(void)
{
  epicsMutexLock(mutexId);
  clientState retval = ourState;
  epicsMutexUnlock(mutexId);
  return(retval);
}
//----------------------------------------------------------------------
const char *infoServ_HTTPClient::getStateText(void)
{
  switch (getState()) {
    case stateInit             : return("stateInit");
    case stateDoingRequest     : return("stateDoingRequest");
    case stateRequestCompleted : return("stateRequestCompleted");
    case stateCleanupFlagged   : return("stateCleanupFlagged");
    case stateThreadExited     : return("stateThreadExited");
  }
  return("unknown");
}
//----------------------------------------------------------------------
infoServ_HTTPServer *infoServ_HTTPClient::getServer()
{
  return(ourServer);
}
//----------------------------------------------------------------------
bool infoServ_HTTPClient::canCleanup()
/*
  Will be called repeatedly by server's main task loop, we use the opportunity
  to check for expiration of recycling timeout since the last request
  completed.

  When we return true server will delete us
*/
{
  epicsMutexLock(mutexId);

  bool retval;
  switch (getState()) {
    case stateInit:
    case stateDoingRequest:
        // can't cleanup until request is completed
        retval = false;
        break;

    case stateRequestCompleted:
        if ((epicsTime::getCurrent() - requestEndTime) > infoServ_threadRecycleSec) {
          // recycle timeout expired
          // flag cleanup, our run() thread task will then exit
          changeState(stateCleanupFlagged);
        }
        retval = false;
        break;

    case stateCleanupFlagged:
        // cleanup has been flagged but thread hasn't exited yet
        retval = false;
        break;

    case stateThreadExited:
        // thread has exited, we can be deleted now
        retval = true;
        break;

    default:
        // shouldn't happen, but something is wrong so we'll croak
        retval = true;
  }


  epicsMutexUnlock(mutexId);
  return(retval);
}
//----------------------------------------------------------------------
bool infoServ_HTTPClient::canRecycle()
{
  return( (getState() == stateRequestCompleted) ? true : false );
}
//----------------------------------------------------------------------
bool infoServ_HTTPClient::recycle(SOCKET socket, const char *desc)
{
  initRequest(socket,desc);
  return(true);
}
//----------------------------------------------------------------------
void infoServ_HTTPClient::start()
{
  thread.start();
}
//----------------------------------------------------------------------
void infoServ_HTTPClient::run()
{
  char tt[TNOW_MAX];

  bool dropOut = false;
  while (! dropOut) {
    // exit thread if our server has stopped
    if (! ourServer->started()) break;

    // otherwise process based on our state
    switch (getState()) {
      case stateInit:
          // shouldn't happen, but let's do something so we don't hog CPU
          epicsThreadSleep(0.1);
          break;

      case stateDoingRequest:
          // repeat trying to read client request until read & handled or we time out
          while (! inputHandled()) {
            if (! ourServer->started()) break;
            if ((epicsTime::getCurrent() - requestStartTime) > infoServ_clientTimeoutSec) {
              printf("%s infoServ: client %s timed out waiting for request\n",tnow(tt),ourDesc);
              break;
            }
          }
          // shutdown our socket
          shutdown(ourSock, SHUT_RDWR);
          // signal that request is done
          requestEndTime = epicsTime::getCurrent();
          changeState(stateRequestCompleted);
          break;

      case stateRequestCompleted:
          // nothing fir us to do but wait to be recycled or cleaned up
          epicsThreadSleep(0.1);
          break;

      case stateCleanupFlagged:
          // time to exit and get cleaned up
          dropOut = true;
          break;

      case stateThreadExited:
          // we've already exited?  okay...
          dropOut = true;
          break;
    } // switch state
  } // while

  // flag that thread has completed
  changeState(stateThreadExited);
}
//----------------------------------------------------------------------
bool infoServ_HTTPClient::inputHandled()
{
  struct timeval timeout;
  fd_set fds;
  char tmpbuff[INTERNAL_BUFFER_SIZE];
  char tt[TNOW_MAX];
  char *pc;
  int rc, kk;

  // see if there's anything to read
  FD_ZERO(&fds);
  FD_SET(ourSock, &fds);
  timeout.tv_sec  = infoServ_selectTimeoutSec;
  timeout.tv_usec = infoServ_selectTimeoutUSec;
  if ((select(ourSock+1, &fds, 0, 0, &timeout) < 1) || (! FD_ISSET(ourSock, &fds))) return(false);

  // data available, read it
  // error or 0 bytes read probably means client closed at their end
  rc = recv(ourSock, tmpbuff, sizeof(tmpbuff)-1, 0);
  if (rc <= 0) return(true);
  tmpbuff[rc] = 0;

  if (infoServ_debugLevel >= 3) {
    printf("client %s read data\n",ourDesc);
    showBuffer(tmpbuff,rc);
  }

  // append to existing input buffer (if there's room)
  kk = strlen(input);
  if ((kk + rc) >= (int)sizeof(input)) {
    int clip = sizeof(input) - kk - 1;
    if (clip >= 0) tmpbuff[clip] = 0;
  }
  strcat(input,tmpbuff);

  // complete HTTP request has blank line at end
  pc = strstr(input,"\r\n\r\n");
  if (! pc) return(false);

  // got complete request
  if (infoServ_debugLevel >= 2) {
    printf("client %s input data\n",ourDesc);
    showBuffer(input,strlen(input));
  }

  // must start with "GET "
  // send 404 error if doesn't
  if (strncmp(input,"GET ",4) != 0) {
    infoServ_send404Error(ourSock,input);
    return(true);
  }
  char *url = input + 4;  // point past GET

  // whack off HTTP/x.x and etc trailer, we ignore it all
  pc = strstr(url," HTTP/");
  if (pc != NULL) (*pc) = 0;

  // convert HTML escape sequences
  infoServ_htmlUnescape(url);

  // store last-requested input
  epicsMutexLock(mutexId);
  strcpy(lastRequest,url);
  if (infoServ_debugLevel > 0)
    printf("%s client %s request '%s'\n",tnow(tt),ourDesc,lastRequest);
  epicsMutexUnlock(mutexId);

  // call server object for handling, it has the url handlers list
  ourServer->handleUrl(ourSock,url);

  // signal we've dealt with request (even if was to send an error to client)
  return(true);
}
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// stuff needed to register global functions with iocsh
//----------------------------------------------------------------------

//----------------------------------------------------------------------
static void iocsh_infoServStart(const iocshArgBuf *args)
{
  infoServStart();
}
static const iocshFuncDef iocsh_infoServStart_FuncDef = {"infoServStart",0,NULL};
//----------------------------------------------------------------------
static void iocsh_infoServStop(const iocshArgBuf *args)
{
  infoServStop();
}
static const iocshFuncDef iocsh_infoServStop_FuncDef = {"infoServStop",0,NULL};
//----------------------------------------------------------------------
static void iocsh_infoServSetParam(const iocshArgBuf *args)
{
  char *param = args[0].sval;
  int value   = args[1].ival;
  infoServSetParam(param,value);
}

static const iocshArg iocsh_infoServSetParam_Arg0 = { "param", iocshArgString };
static const iocshArg iocsh_infoServSetParam_Arg1 = { "value", iocshArgInt };
static const iocshArg * const iocsh_infoServSetParam_Args[2] =
{&iocsh_infoServSetParam_Arg0, &iocsh_infoServSetParam_Arg1};
static const iocshFuncDef iocsh_infoServSetParam_FuncDef = {"infoServSetParam",2,iocsh_infoServSetParam_Args};

/*--------------------------------------------------------------------*/
static void iocsh_infoServAllowFilePath(const iocshArgBuf *args)
{
  char *path = args[0].sval;
  infoServAllowFilePath(path);
}

static const iocshArg iocsh_infoServAllowFilePath_Arg0 = { "path", iocshArgString };
static const iocshArg * const iocsh_infoServAllowFilePath_Args[1] =
{&iocsh_infoServAllowFilePath_Arg0};
static const iocshFuncDef iocsh_infoServAllowFilePath_FuncDef = {"infoServAllowFilePath",1,iocsh_infoServAllowFilePath_Args};

/*--------------------------------------------------------------------*/
static void iocsh_infoServReport(const iocshArgBuf *args)
{
  int level = args[0].ival;

  infoServReport(level);
}

static const iocshArg iocsh_infoServReport_Arg0 = { "level", iocshArgInt };
static const iocshArg * const iocsh_infoServReport_Args[1] = {&iocsh_infoServReport_Arg0};
static const iocshFuncDef iocsh_infoServReport_FuncDef = {"infoServReport",1,iocsh_infoServReport_Args};

/*--------------------------------------------------------------------*/
// set up registrar function
//----------------------------------------------------------------------
static void infoServ_Registrar(void)
{
  // register init hook function
  initHookRegister(infoServStartHookFunc);

#ifdef HAS_AT_EXIT
  // register at-exit call
  long stat = epicsAtExit(infoServAtExit,NULL);
  if (stat != epicsOk) {
    printf("infoServ: epicsAtExit returned %ld\n",stat);
  }
#endif

  // create server object
  // (thread won't start until infoServStart() is called */
  requireServerObject();

  // register shell commands
  iocshRegister(&iocsh_infoServStart_FuncDef,         iocsh_infoServStart);
  iocshRegister(&iocsh_infoServStop_FuncDef,          iocsh_infoServStop);
  iocshRegister(&iocsh_infoServSetParam_FuncDef,      iocsh_infoServSetParam);
  iocshRegister(&iocsh_infoServAllowFilePath_FuncDef, iocsh_infoServAllowFilePath);
  iocshRegister(&iocsh_infoServReport_FuncDef,        iocsh_infoServReport);
}
extern "C" {
  epicsExportRegistrar(infoServ_Registrar);
}
/*--------------------------------------------------------------------*/


