/*
  infoServ.h
*/

#ifndef _infoServH_
#define _infoServH_

#include <epicsThread.h>
#include <epicsMutex.h>
#include <epicsTime.h>
#include <osiSock.h>
#include <dbStaticLib.h>
#include <ellLib.h>

#define DEFAULT_INFOSERV_PORT 7729

typedef struct {
  long upSeconds;
  long currentClientCount;
  long highestClientCount;
  long createdClientCount;
  long recycledClientCount;
  long deletedClientCount;
} infoServ_serverStats;

//----------------------------------------------------------------------
class infoServ_requestHandler
// virtual base class, descendants must override pure virtual methods
{
  public:
    infoServ_requestHandler();
    virtual ~infoServ_requestHandler();
    virtual bool canHandleUrl(const char *urlText) = 0;
    virtual bool handleUrl(SOCKET sock, const char *urlText) = 0;
    virtual const char *htmlHelpTableRows();
    virtual const char *infoString();
    virtual void serverStarted();
    virtual void serverStopped();
    virtual void report(int level);
  protected:
    virtual void sendTextHeader(SOCKET sock, const char *extraHeaderLines = NULL);
    virtual void sendJSONHeader(SOCKET sock, const char *extraHeaderLines = NULL);
};

//----------------------------------------------------------------------
class infoServ_HTTPServer : public epicsThreadRunable
{
  public:
    infoServ_HTTPServer(short port);
    virtual ~infoServ_HTTPServer();

    void start();
    void stop();
    bool started();

    void addAllowedFilePath(const char *newpath);
    void showAllowedFilePaths(void);
    bool isAllowedFilePath(const char *filename);

    void changePort(short newport);
    void report(int level);

    bool registerUrlHandler(infoServ_requestHandler *handler);
    bool unregisterUrlHandler(infoServ_requestHandler *handler);

    void getServerStats(infoServ_serverStats *stats);
    void getAllowedFilePaths(ELLLIST *L);
    void getHandlerHelpTableRows(ELLLIST *L);
    void getHandlerInfoStrings(ELLLIST *L);

    void handleUrl(SOCKET sock, const char *url);

  protected:
    epicsThread          *pThread;
    bool                  stopThread;
    bool                  needCreateSocket;
    bool                  isStarted;
    short                 ourPort;
    SOCKET                ourSock;
    infoServ_serverStats  serverStats;
    epicsMutexId          listsMutexId;
    ELLLIST               clientList;
    ELLLIST               allowedFilePathsList;
    ELLLIST               urlHandlerList;
    epicsTime             startTime;

    void run();
    int cleanup();
    bool assignClientThread(SOCKET clientsock, const char *desc);
};

//----------------------------------------------------------------------

#define CLIENT_INBUFF_SIZE  1024

class infoServ_HTTPClient : public epicsThreadRunable
{
  public:
    infoServ_HTTPClient(infoServ_HTTPServer *server, SOCKET socket, const char *desc);
    virtual ~infoServ_HTTPClient();
    infoServ_HTTPServer *getServer();
    bool canCleanup();
    bool canRecycle();
    bool recycle(SOCKET socket, const char *desc);
    void start();
    void run();
    void report(int level);

    enum clientState {
      stateInit             = 0,
      stateDoingRequest     = 1,
      stateRequestCompleted = 2,
      stateCleanupFlagged   = 3,
      stateThreadExited     = 4
    };

  protected:
    epicsThread          thread;
    infoServ_HTTPServer *ourServer;
    epicsTime            requestStartTime;
    epicsTime            requestEndTime;
    char                *ourDesc;
    clientState          ourState;
    SOCKET               ourSock;
    epicsMutexId         mutexId;
    char                 input[CLIENT_INBUFF_SIZE];
    char                 lastRequest[CLIENT_INBUFF_SIZE];

    const char *getStartupFile(void);
    bool inputHandled();
    void initRequest(SOCKET socket, const char *desc);
    void changeState(clientState newState);
    infoServ_HTTPClient::clientState getState(void);
    const char *getStateText(void);
};

//----------------------------------------------------------------------
class infoServ_recordData
{
  public:
    DBENTRY *hog;
    infoServ_recordData(DBENTRY *pdbentry);
    virtual ~infoServ_recordData(void);

    const char *recordName(void);
    const char *recordType(void);
    bool isAlias(void);
    const char *aliasTo(void);
    ELLLIST *aliasesOf(void);

    void addAliasOf(const char *other);

  protected:
    char    *name;
    char    *rtyp;
    char    *alias_to;
    DBENTRY *entry;
    ELLLIST  aliases_of;
};

//----------------------------------------------------------------------

typedef struct {
  ELLNODE node;
  char *text;
} ELL_textNode;

typedef struct {
  ELLNODE node;
  infoServ_recordData *record;
} ELL_recordDataNode;

typedef int (*ELL_COMPARE_FUNC)(ELLNODE *n1, ELLNODE *n2);

//
// callable from iocsh (and other code)
//
void infoServStart(void);
void infoServStop(void);
void infoServAllowFilePath(const char *newpath);
int infoServSetParam(const char *param, int value);
void infoServReport(int level);

//
// callable from other code
//
infoServ_HTTPServer *infoServ_getServer(void);

bool infoServ_registerUrlHandler(infoServ_requestHandler *handler);
bool infoServ_unregisterUrlHandler(infoServ_requestHandler *handler);

const char *infoServ_getStartupFile(char *buffer, int maxlen);

void infoServ_parseCommaSeparated(const char *str, ELLLIST *L, const char *endChars = " &");
void infoServ_copyTextList(ELLLIST *targ, ELLLIST *src);
void infoServ_destroyTextList(ELLLIST *L);

void infoServ_buildRecordList(ELLLIST *L);
void infoServ_destroyRecordList(ELLLIST *L);

bool infoServ_sortList(ELLLIST *L, ELL_COMPARE_FUNC func);
int infoServ_sortCompareTextNodes(ELLNODE *n1, ELLNODE *n2);
int infoServ_sortCompareRecordNodeNames(ELLNODE *n1, ELLNODE *n2);

void infoServ_htmlUnescape(char *buff);
bool infoServ_needJSONEscape(const char *buff);
int infoServ_JSONEscapeSize(const char *buff);
void infoServ_JSONEscape(char *buff);

void infoServ_socketSendString(SOCKET sock, const char *str);
void infoServ_sendHTTPText(SOCKET sock, const char *statcode, const char *contents);
void infoServ_send404Error(SOCKET sock, const char *badurl);
void infoServ_send503Error(SOCKET sock);
void infoServ_socketSendJSONText(SOCKET sock, const char *text);
void infoServ_socketSendJSONFieldName(SOCKET sock, const char *fieldname);
void infoServ_socketSendJSONPairText(SOCKET sock, const char *fieldname, const char *fieldvalue);
void infoServ_socketSendJSONPairLong(SOCKET sock, const char *fieldname, long fieldvalue);
void infoServ_socketSendJSONPairBool(SOCKET sock, const char *fieldname, bool fieldvalue);
void infoServ_socketSendJSONPairDouble(SOCKET sock, const char *fieldname, const char *format, double fieldvalue);

#endif /* _infoServH_ */


