/*
  testAddIn.cpp

  Demonstration add-in for infoServ

  Handlers two URLs:
    /bob? : returns some demonstration JSON stuff
    /tim? : returns some demonstration text

  2015-04-03  J. Priller  send HTTP headers with replies
  2015-03-17  J. Priller  original version
*/

#include <stdio.h>
#include <string.h>
#include <epicsVersion.h>
#include <epicsExport.h>
#include <iocsh.h>

#include "infoServ.h"

#ifndef epicsError
  #define epicsError -1
#endif
#ifndef epicsOk
  #define epicsOk 0
#endif

//----------------------------------------------------------------------
// internal class definitions
//----------------------------------------------------------------------
class testAddIn_requestHandler : public infoServ_requestHandler
{
  public:
    testAddIn_requestHandler();
    virtual ~testAddIn_requestHandler();
    virtual bool canHandleUrl(const char *urlText);
    virtual bool handleUrl(SOCKET sock, const char *urlText);
    virtual const char *htmlHelpTableRows();
    virtual const char *infoString();
    virtual void serverStarted();
    virtual void serverStopped();
    virtual void report(int level);
};

//----------------------------------------------------------------------
// forward declarations
//----------------------------------------------------------------------
void testAddInStart(void);
void testAddInStop(void);

//----------------------------------------------------------------------
// module globals
//----------------------------------------------------------------------

static testAddIn_requestHandler *ourHandler = NULL;

//----------------------------------------------------------------------
// iocsh callable functions
//----------------------------------------------------------------------
void testAddInStart(void)
{
  // if our handler isn't already created, create and register it

  if (ourHandler) {
    printf("testAddInStart was already called\n");
    return;
  }

  // create handler object
  ourHandler = new testAddIn_requestHandler();

  // try to register it with infoServ
  if (! infoServ_registerUrlHandler(ourHandler)) {
    printf("ERROR: infoServ_registerUrlHandler FAILED\n");
    delete ourHandler;
    ourHandler = NULL;
    return;
  }

  // success if we get here
  printf("testAddIn started\n");
}
//----------------------------------------------------------------------
void testAddInStop(void)
{
  // if we have a handler object created, de-register it and delete it

  // de-register url handler
  if (! ourHandler) {
    printf("testAddIn had not been started\n");
    return;
  }

  if (infoServ_unregisterUrlHandler(ourHandler)) {
    // unregister succeeded, we can delete object
    delete ourHandler;
  } else {
    // if unregister does NOT succeed then server already deleted object
    // (maybe IOC is shutting down)
    // there's nothing else we can or should do about it
  }
  ourHandler = NULL;

  // show that we've heard and obeyed
  printf("testAddIn stopped\n");
}
//----------------------------------------------------------------------
// class testAddIn_requestHandler
//----------------------------------------------------------------------
testAddIn_requestHandler::testAddIn_requestHandler()
  : infoServ_requestHandler()
{
  /*
    Nothing we need to do here

    If we were a more complicated example we could allocate memory
    and initialize data structures, etc
  */
  printf("testAddIn object created\n");
}
//----------------------------------------------------------------------
testAddIn_requestHandler::~testAddIn_requestHandler()
{
  /*
    Nothing we need to do here

    If we were a more complicated example we'd release memory
    we'd allocated previously and etc
  */
  printf("testAddIn object destroyed\n");
}
//----------------------------------------------------------------------
bool testAddIn_requestHandler::canHandleUrl(const char *urlText)
{
  /*
    infoServ is asking us if we can handle the given trailing URL text
    ( in 'http://host:port/whatever' this would be '/whatever' )

    We return true if this is something we handle and false otherwise
  */
  if (strncmp(urlText,"/bob?",5) == 0) return(true);
  if (strncmp(urlText,"/tim?",5) == 0) return(true);
  return(false);
}
//----------------------------------------------------------------------
bool testAddIn_requestHandler::handleUrl(SOCKET sock, const char *urlText)
{
  /*
    infoServ is asking us to handle the given trailing URL text
    ( in 'http://host:port/whatever' this would be '/whatever' )

    Here a real URL handler would parse the URL for what the user is asking for
    and send back the pertinent information.

    We're just a demo, we just send back some demonstration content.

    If for some reason we can't handle the URL (maybe a memory alloc failed?)
    we can either return 'false' and infoServ will send the user a 404 Not Found
    response, or we can send the user a custom error message in text or in HTML
    if that's what we wanted to do, and then return 'true' to signal we handled
    the user's request.
  */

  if (strncmp(urlText,"/bob?",5) == 0) {
    // use infoServ convenience functions to send some JSON stuff
    // note that we're responsible for making sure commas and braces/brackets are correct
    sendJSONHeader(sock);

    infoServ_socketSendString(sock,"{ \"bob\": {\r\n");
    infoServ_socketSendJSONPairText(sock,"urlText",urlText);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairText(sock,"text","this is some text");

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairLong(sock,"long",929560);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairBool(sock,"bool_false",false);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairBool(sock,"bool_true",true);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairDouble(sock,"double_e","%e",3.1415926);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairDouble(sock,"double_f","%f",2.7182818);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONPairDouble(sock,"double_g","%g",6.0221413e+23);

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONFieldName(sock,"a_list");
    infoServ_socketSendString(sock,"[ ");
    infoServ_socketSendJSONText(sock,"text");
    infoServ_socketSendString(sock,", 12345");
    infoServ_socketSendString(sock,", true");
    infoServ_socketSendString(sock,", 98.765");
    infoServ_socketSendString(sock," ]");

    infoServ_socketSendString(sock,",\r\n");
    infoServ_socketSendJSONFieldName(sock,"a_dict");
    infoServ_socketSendString(sock,"{ ");
    infoServ_socketSendJSONPairText(sock,"something","else");
    infoServ_socketSendString(sock,", ");
    infoServ_socketSendJSONPairLong(sock,"fnords",17);
    infoServ_socketSendString(sock,", ");
    infoServ_socketSendJSONPairBool(sock,"is_it_soup_yet",false);
    infoServ_socketSendString(sock,", ");
    infoServ_socketSendJSONPairDouble(sock,"furlongs","%g",11.2);
    infoServ_socketSendString(sock," }");

    infoServ_socketSendString(sock," }\r\n");
    infoServ_socketSendString(sock,"}\r\n");
    return(true);
  }

  if (strncmp(urlText,"/tim?",5) == 0) {
    // send some text
    sendTextHeader(sock);
    infoServ_socketSendString(sock,"this is a test of the 'bob' infoServ URL\r\n");
    infoServ_socketSendString(sock,"the URL passed to it was:\r\n");
    infoServ_socketSendString(sock,urlText);
    return(true);
  }

  // if we get here we got called in error
  return(false);
}
//----------------------------------------------------------------------
const char *testAddIn_requestHandler::htmlHelpTableRows()
{
  /*
    Use this function to return the HTML row content for infoServ to use in its
    /help? table of available URLs.

    If you don't want your handler to appear in /help?, then either don't
    override the htmlHelpTableRows() method in your objects or return NULL.
  */
  const char *helpRows = \
    "<tr>\r\n"                                                       \
    "  <td><b>/bob?</b></td>\r\n"                                    \
    "  <td>(JSON data) testAddIn demo</td>\r\n"                      \
    "  <td align=\"center\"><a href=\"/bob?\">test it!</a></td>\r\n" \
    "</tr>\r\n"                                                      \
    "<tr>\r\n"                                                       \
    "  <td><b>/tim?</b></td>\r\n"                                    \
    "  <td>(text data) testAddIn demo</td>\r\n"                      \
    "  <td align=\"center\"><a href=\"/tim?\">test it!</a></td>\r\n" \
    "</tr>\r\n";
  return(helpRows);
}
//----------------------------------------------------------------------
const char *testAddIn_requestHandler::infoString()
{
  /*
    Use this function to return a short informational string to be included
    in infoServ's /info? reply where it lists known url-handlers.

    If you don't want your handler to appear in /info?, then either don't
    override the infoString() method in your objects or return NULL.
  */
  return( "/bob?, /tim? (testAddIn)" );
}
//----------------------------------------------------------------------
void testAddIn_requestHandler::serverStarted()
{
  /*
    This method is called whenever infoServ is (re)started after
    we've registered ourselves with it.

    We don't need to do anything, though we could if there was some reason to.
  */
  printf("testAddIn sees that the infoServ service was (re)started\n");
}
//----------------------------------------------------------------------
void testAddIn_requestHandler::serverStopped()
{
  /*
    This method is called whenever infoServ is stopped after
    we've registered ourselves with it.

    We don't need to do anything, we'll stay registered and be informed if/when
    infoServ is restarted by a call to our serverStarted() method.
  */
  printf("testAddIn sees that the infoServ service was stopped\n");
}
//----------------------------------------------------------------------
void testAddIn_requestHandler::report(int level)
{
  /*
    This method is called when infoServReport is called from the IOC shell.

    We can report any internal details we'd like, but as we're just a demo
    there's nothing much to say.
  */
  printf("testAddIn report called with level %d\n",level);
}

//----------------------------------------------------------------------
// stuff needed to register global functions with iocsh
//----------------------------------------------------------------------

//----------------------------------------------------------------------
static void iocsh_testAddInStart(const iocshArgBuf *args)
{
  testAddInStart();
}
static const iocshFuncDef iocsh_testAddInStart_FuncDef = {"testAddInStart",0,NULL};
//----------------------------------------------------------------------
static void iocsh_testAddInStop(const iocshArgBuf *args)
{
  testAddInStop();
}
static const iocshFuncDef iocsh_testAddInStop_FuncDef = {"testAddInStop",0,NULL};
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// set up registrar function
//----------------------------------------------------------------------
static void testAddIn_Registrar(void)
{
  // register shell commands
  iocshRegister(&iocsh_testAddInStart_FuncDef,         iocsh_testAddInStart);
  iocshRegister(&iocsh_testAddInStop_FuncDef,          iocsh_testAddInStop);
}
extern "C" {
  epicsExportRegistrar(testAddIn_Registrar);
}
/*--------------------------------------------------------------------*/


